import json
import numpy as np
import os
import pandas as pd
from pathlib import Path
import tsdf
from typing import List

from paradigma.classification import ClassifierPackage
from paradigma.config import IMUConfig, PPGConfig, GaitConfig, TremorConfig, PulseRateConfig
from paradigma.constants import DataColumns, TimeUnit
from paradigma.pipelines.gait_pipeline import extract_gait_features, detect_gait, \
    extract_arm_activity_features, filter_gait
from paradigma.pipelines.tremor_pipeline import extract_tremor_features, detect_tremor, \
    aggregate_tremor
from paradigma.pipelines.pulse_rate_pipeline import extract_signal_quality_features, signal_quality_classification, \
    aggregate_pulse_rate
from paradigma.preprocessing import preprocess_imu_data, preprocess_ppg_data
from paradigma.util import read_metadata, write_df_data, get_end_iso8601, merge_predictions_with_timestamps


def preprocess_imu_data_io(path_to_input: str | Path, path_to_output: str | Path, 
                           config: IMUConfig, sensor: str, watch_side: str) -> None:
    # Load data
    metadata_time, metadata_values = read_metadata(str(path_to_input), str(config.meta_filename),
                                                    str(config.time_filename), str(config.values_filename))
    df = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    # Preprocess data
    df = preprocess_imu_data(df=df, config=config, sensor=sensor, watch_side=watch_side)

    # Store data
    for sensor, units in zip(['accelerometer', 'gyroscope'], ['g', config.rotation_units]):
        if any(sensor in col for col in df.columns):
            df_sensor = df[[DataColumns.TIME] + [x for x in df.columns if sensor in x]]

            metadata_values.channels = [x for x in df.columns if sensor in x]
            metadata_values.units = list(np.repeat(units, len(metadata_values.channels)))
            metadata_values.scale_factors = []
            metadata_values.file_name = f'{sensor}_values.bin'

            metadata_time.file_name = f'{sensor}_time.bin'
            metadata_time.units = [TimeUnit.RELATIVE_S]

            write_df_data(metadata_time, metadata_values, path_to_output, f'{sensor}_meta.json', df_sensor)


def preprocess_ppg_data_io(path_to_input_ppg: str | Path, path_to_input_imu: str | Path,
                           output_path: str | Path, ppg_config: PPGConfig, 
                           imu_config: IMUConfig) -> None:
    """	
    Preprocess PPG and IMU data by resampling, filtering, and aligning the data segments.

    Parameters
    ----------
    path_to_input_ppg : str | Path
        Path to the PPG data.
    path_to_input_imu : str | Path
        Path to the IMU data.
    output_path : str | Path
        Path to store the preprocessed data.
    ppg_config : PPGConfig
        Configuration object for PPG preprocessing.
    imu_config : IMUConfig
        Configuration object for IMU preprocessing.

    Returns
    -------
    None
    """ 

    # Load PPG data
        # Load data
    metadata_time_ppg, metadata_values_ppg = read_metadata(path_to_input_ppg, ppg_config.meta_filename,
                                                    ppg_config.time_filename, ppg_config.values_filename)
    df_ppg = tsdf.load_dataframe_from_binaries([metadata_time_ppg, metadata_values_ppg], tsdf.constants.ConcatenationType.columns)

    # Load IMU data
    metadata_time_imu, metadata_values_imu = read_metadata(path_to_input_imu, imu_config.meta_filename,
                                                    imu_config.time_filename, imu_config.values_filename)
    df_imu = tsdf.load_dataframe_from_binaries([metadata_time_imu, metadata_values_imu], tsdf.constants.ConcatenationType.columns)

    # Drop the gyroscope columns from the IMU data
    cols_to_drop = df_imu.filter(regex='^gyroscope_').columns
    df_acc = df_imu.drop(cols_to_drop, axis=1)

    # Preprocess data
    df_ppg_proc, df_acc_proc = preprocess_ppg_data(
        df_ppg=df_ppg, 
        df_acc=df_acc, 
        ppg_config=ppg_config, 
        imu_config=imu_config,
        start_time_ppg=metadata_time_ppg.start_iso8601,
        start_time_imu=metadata_time_imu.start_iso8601
    )

    # Store data
    metadata_values_imu.channels = list(imu_config.d_channels_accelerometer.keys())
    metadata_values_imu.units = list(imu_config.d_channels_accelerometer.values())
    metadata_values_imu.file_name = 'accelerometer_values.bin'
    metadata_time_imu.units = [TimeUnit.ABSOLUTE_MS]
    metadata_time_imu.file_name = 'accelerometer_time.bin'
    write_df_data(metadata_time_imu, metadata_values_imu, output_path, 'accelerometer_meta.json', df_acc_proc)

    metadata_values_ppg.channels = list(ppg_config.d_channels_ppg.keys())
    metadata_values_ppg.units = list(ppg_config.d_channels_ppg.values())
    metadata_values_ppg.file_name = 'PPG_values.bin'
    metadata_time_ppg.units = [TimeUnit.ABSOLUTE_MS]
    metadata_time_ppg.file_name = 'PPG_time.bin'
    write_df_data(metadata_time_ppg, metadata_values_ppg, output_path, 'PPG_meta.json', df_ppg_proc)


def extract_gait_features_io(
        config: GaitConfig,
        path_to_input: str | Path, 
        path_to_output: str | Path
    ) -> None:
    # Load data
    metadata_time, metadata_values = read_metadata(path_to_input, config.meta_filename, config.time_filename, config.values_filename)
    df = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    # Extract gait features
    df_features = extract_gait_features(df=df, config=config)

    # Store data
    end_iso8601 = get_end_iso8601(start_iso8601=metadata_time.start_iso8601,
                                  window_length_seconds=int(df_features[DataColumns.TIME][-1:].values[0] + config.window_length_s))

    metadata_values.file_name = 'gait_values.bin'
    metadata_time.file_name = 'gait_time.bin'
    metadata_values.end_iso8601 = end_iso8601
    metadata_time.end_iso8601 = end_iso8601
    
    metadata_values.channels = list(config.d_channels_values.keys())
    metadata_values.units = list(config.d_channels_values.values())

    metadata_time.channels = [DataColumns.TIME]
    metadata_time.units = [TimeUnit.RELATIVE_S]

    write_df_data(metadata_time, metadata_values, path_to_output, 'gait_meta.json', df_features)


def detect_gait_io(
        config: GaitConfig, 
        path_to_input: str | Path, 
        path_to_output: str | Path, 
        full_path_to_classifier_package: str | Path, 
    ) -> None:
    
    # Load the data
    config.set_filenames('gait')

    metadata_time, metadata_values = read_metadata(path_to_input, config.meta_filename, config.time_filename, config.values_filename)
    df = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    clf_package = ClassifierPackage.load(full_path_to_classifier_package)

    df[DataColumns.PRED_GAIT_PROBA] = detect_gait(
        df=df, 
        clf_package=clf_package
    )

    # Prepare the metadata
    metadata_values.file_name = 'gait_values.bin'
    metadata_time.file_name = 'gait_time.bin'

    metadata_values.channels = [DataColumns.PRED_GAIT_PROBA]
    metadata_values.units = ['probability']

    metadata_time.channels = [DataColumns.TIME]
    metadata_time.units = [TimeUnit.RELATIVE_S]

    write_df_data(metadata_time, metadata_values, path_to_output, 'gait_meta.json', df)


def extract_arm_activity_features_io(
        config: GaitConfig, 
        path_to_timestamp_input: str | Path, 
        path_to_prediction_input: str | Path, 
        full_path_to_classifier_package: str | Path, 
        path_to_output: str | Path
    ) -> None:
    # Load accelerometer and gyroscope data
    dfs = []
    for sensor in ['accelerometer', 'gyroscope']:
        config.set_sensor(sensor)
        meta_ts_filename = f'{sensor}_meta.json'
        values_ts_filename = f'{sensor}_values.bin'
        time_ts_filename = f'{sensor}_time.bin'

        metadata_ts_dict = tsdf.load_metadata_from_path(os.path.join(path_to_timestamp_input, meta_ts_filename))
        metadata_ts_time = metadata_ts_dict[time_ts_filename]
        metadata_ts_values = metadata_ts_dict[values_ts_filename]
        dfs.append(tsdf.load_dataframe_from_binaries([metadata_ts_time, metadata_ts_values], tsdf.constants.ConcatenationType.columns))

    df_ts = pd.merge(dfs[0], dfs[1], on=DataColumns.TIME)

    # Load gait predictions
    meta_pred_filename = 'gait_meta.json'
    values_pred_filename = 'gait_values.bin'
    time_pred_filename = 'gait_time.bin'

    metadata_pred_dict = tsdf.load_metadata_from_path(os.path.join(path_to_prediction_input, meta_pred_filename))
    metadata_pred_time = metadata_pred_dict[time_pred_filename]
    metadata_pred_values = metadata_pred_dict[values_pred_filename]

    df_pred_gait = tsdf.load_dataframe_from_binaries([metadata_pred_time, metadata_pred_values], tsdf.constants.ConcatenationType.columns)

    clf_package = ClassifierPackage.load(full_path_to_classifier_package)

    gait_preprocessing_config = GaitConfig(step='gait')

    df = merge_predictions_with_timestamps(
        df_ts=df_ts, 
        df_predictions=df_pred_gait, 
        pred_proba_colname=DataColumns.PRED_GAIT_PROBA,
        window_length_s=gait_preprocessing_config.window_length_s,
        fs=gait_preprocessing_config.sampling_frequency
    )

    # Add a column for predicted gait based on a fitted threshold
    df[DataColumns.PRED_GAIT] = (df[DataColumns.PRED_GAIT_PROBA] >= clf_package.threshold).astype(int)

    # Filter the DataFrame to only include predicted gait (1)
    df = df.loc[df[DataColumns.PRED_GAIT]==1].reset_index(drop=True)

    # Extract arm activity features
    config = GaitConfig(step='arm_activity')
    df_features = extract_arm_activity_features(
        df=df, 
        config=config,
    )

    end_iso8601 = get_end_iso8601(metadata_ts_values.start_iso8601, df_features[DataColumns.TIME][-1:].values[0] + config.window_length_s)

    metadata_ts_values.end_iso8601 = end_iso8601
    metadata_ts_values.file_name = 'arm_activity_values.bin'
    metadata_ts_time.end_iso8601 = end_iso8601
    metadata_ts_time.file_name = 'arm_activity_time.bin'

    metadata_ts_values.channels = list(config.d_channels_values.keys())
    metadata_ts_values.units = list(config.d_channels_values.values())

    metadata_ts_time.channels = [DataColumns.TIME]
    metadata_ts_time.units = [TimeUnit.RELATIVE_S]

    write_df_data(metadata_ts_time, metadata_ts_values, path_to_output, 'arm_activity_meta.json', df_features)


def filter_gait_io(
        config: GaitConfig, 
        path_to_input: str | Path, 
        path_to_output: str | Path, 
        full_path_to_classifier_package: str | Path, 
    ) -> None:
    # Load the data
    config.set_filenames('arm_activity')

    metadata_time, metadata_values = read_metadata(path_to_input, config.meta_filename, config.time_filename, config.values_filename)
    df = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    clf_package = ClassifierPackage.load(filepath=full_path_to_classifier_package)

    df[DataColumns.PRED_NO_OTHER_ARM_ACTIVITY_PROBA] = filter_gait(
        df=df, 
        clf_package=clf_package
    )

    # Prepare the metadata
    metadata_values.file_name = 'arm_activity_values.bin'
    metadata_time.file_name = 'arm_activity_time.bin'

    metadata_values.channels = [DataColumns.PRED_NO_OTHER_ARM_ACTIVITY_PROBA]
    metadata_values.units = ['probability']

    metadata_time.channels = [DataColumns.TIME]
    metadata_time.units = [TimeUnit.RELATIVE_S]

    write_df_data(metadata_time, metadata_values, path_to_output, 'arm_activity_meta.json', df)


def extract_tremor_features_io(input_path: str | Path, output_path: str | Path, config: TremorConfig) -> None:
    # Load data
    metadata_time, metadata_values = read_metadata(input_path, config.meta_filename, config.time_filename, config.values_filename)
    df = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    # Extract tremor features
    df_windowed = extract_tremor_features(df, config)

    # Store data
    end_iso8601 = get_end_iso8601(start_iso8601=metadata_time.start_iso8601,
                                  window_length_seconds=int(df_windowed[DataColumns.TIME][-1:].values[0] + config.window_length_s))

    metadata_values.end_iso8601 = end_iso8601
    metadata_values.file_name = 'tremor_values.bin'
    metadata_time.end_iso8601 = end_iso8601
    metadata_time.file_name = 'tremor_time.bin'

    metadata_values.channels = list(config.d_channels_values.keys())
    metadata_values.units = list(config.d_channels_values.values())

    metadata_time.channels = [DataColumns.TIME]
    metadata_time.units = ['relative_time_ms']

    write_df_data(metadata_time, metadata_values, output_path, 'tremor_meta.json', df_windowed)


def detect_tremor_io(input_path: str | Path, output_path: str | Path, path_to_classifier_input: str | Path, config: TremorConfig) -> None:
    
    # Load the data
    config.set_filenames('tremor')

    metadata_time, metadata_values = read_metadata(input_path, config.meta_filename, config.time_filename, config.values_filename)
    df = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    df = detect_tremor(df, config, path_to_classifier_input)

    # Prepare the metadata
    metadata_values.file_name = 'tremor_values.bin'
    metadata_time.file_name = 'tremor_time.bin'

    metadata_values.channels = list(config.d_channels_values.keys())
    metadata_values.units = list(config.d_channels_values.values())

    metadata_time.channels = [DataColumns.TIME]
    metadata_time.units = ['relative_time_ms']

    write_df_data(metadata_time, metadata_values, output_path, 'tremor_meta.json', df)


def aggregate_tremor_io(path_to_feature_input: str | Path, path_to_prediction_input: str | Path, output_path: str | Path, config: TremorConfig) -> None:
    
    # Load the features & predictions
    metadata_time, metadata_values = read_metadata(path_to_feature_input, config.meta_filename, config.time_filename, config.values_filename)
    df_features = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    metadata_dict = tsdf.load_metadata_from_path(path_to_prediction_input / config.meta_filename)
    metadata_time = metadata_dict[config.time_filename]
    metadata_values = metadata_dict[config.values_filename]
    df_predictions = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    # Subset features
    df_features = df_features[['tremor_power', 'below_tremor_power']]

    # Concatenate predictions and tremor power
    df = pd.concat([df_predictions, df_features], axis=1)

    # Compute aggregated tremor measures
    d_aggregates = aggregate_tremor(df, config)

    # Save output
    with open(output_path / "tremor_aggregates.json", 'w') as json_file:
        json.dump(d_aggregates, json_file, indent=4)


def extract_signal_quality_features_io(input_path: str | Path, output_path: str | Path, ppg_config: PulseRateConfig, acc_config: PulseRateConfig) -> pd.DataFrame:
    """
    Extract signal quality features from the PPG signal and save them to a file.

    Parameters
    ----------
    input_path : str | Path
        The path to the directory containing the preprocessed PPG and accelerometer data.
    output_path : str | Path
        The path to the directory where the extracted features will be saved.
    ppg_config: PulseRateConfig
        The configuration for the signal quality feature extraction of the ppg signal.
    acc_config: PulseRateConfig
        The configuration for the signal quality feature extraction of the accelerometer signal.

    Returns
    -------
    df_windowed : pd.DataFrame
        The DataFrame containing the extracted signal quality features.

    """	
    # Load PPG data
    metadata_time, metadata_values = read_metadata(input_path, ppg_config.meta_filename, ppg_config.time_filename, ppg_config.values_filename)
    df_ppg = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)
    
    # Load IMU data
    metadata_time, metadata_values = read_metadata(input_path, acc_config.meta_filename, acc_config.time_filename, acc_config.values_filename)
    df_acc = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    # Extract signal quality features
    df_windowed = extract_signal_quality_features(df_ppg, df_acc, ppg_config, acc_config)
    
    # Save the extracted features
    #TO BE ADDED
    return df_windowed


def signal_quality_classification_io(input_path: str | Path, output_path: str | Path, path_to_classifier_input: str | Path, config: PulseRateConfig) -> None:
    
    # Load the data
    metadata_time, metadata_values = read_metadata(input_path, config.meta_filename, config.time_filename, config.values_filename)
    df_windowed = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    df_sqa = signal_quality_classification(df_windowed, config, path_to_classifier_input)


def aggregate_pulse_rate_io(
        full_path_to_input: str | Path, 
        full_path_to_output: str | Path, 
        aggregates: List[str] = ['mode', '99p']
    ) -> None:
    """
    Extract pulse rate from the PPG signal and save the aggregated pulse rate estimates to a file.

    Parameters
    ----------
    input_path : str | Path
        The path to the directory containing the pulse rate estimates.
    output_path : str | Path
        The path to the directory where the aggregated pulse rate estimates will be saved.
    aggregates : List[str]
        The list of aggregation methods to be used for the pulse rate estimates. The default is ['mode', '99p'].
    """

    # Load the pulse rate estimates
    with open(full_path_to_input, 'r') as f:
        df_pr = json.load(f)
    
    # Aggregate the pulse rate estimates
    pr_values = df_pr['pulse_rate'].values
    df_pr_aggregates = aggregate_pulse_rate(pr_values, aggregates)

    # Save the aggregated pulse rate estimates
    with open(full_path_to_output, 'w') as json_file:
        json.dump(df_pr_aggregates, json_file, indent=4)


import numpy as np
from scipy import signal
from typing import Tuple

from paradigma.config import PulseRateConfig


def assign_sqa_label(
        ppg_prob: np.ndarray, 
        config: PulseRateConfig, 
        acc_label=None
    ) -> np.ndarray:
    """
    Assigns a signal quality label to every individual data point.

    Parameters
    ----------
    ppg_prob : np.ndarray
        The probabilities for PPG.
    config : PulseRateConfig
        The configuration parameters.
    acc_label : np.ndarray, optional
        The labels for the accelerometer.

    Returns
    -------
    np.ndarray
        The signal quality assessment labels.
    """
    # Default _label to ones if not provided
    if acc_label is None:
        acc_label = np.ones(len(ppg_prob))

    # Number of samples in an epoch
    fs = config.sampling_frequency
    samples_per_epoch = config.window_length_s * fs

    # Calculate number of samples to shift for each epoch
    samples_shift = config.window_step_length_s * fs
    n_samples = int(np.round(len(ppg_prob) + config.window_overlap_s) * fs)
    data_prob = np.zeros(n_samples)
    data_label_imu = np.zeros(n_samples, dtype=np.int8)

    for i in range(n_samples):
        # Start and end indices for current epoch
        start_idx = max(0, int((i - (samples_per_epoch - samples_shift)) // fs)) # max to handle first epochs
        end_idx = min(int(i // fs), len(ppg_prob))  # min to handle last epochs  

        # Extract probabilities and labels for the current epoch
        prob = ppg_prob[start_idx:end_idx+1]
        label_imu = acc_label[start_idx:end_idx+1]

        # Calculate mean probability and majority voting for labels
        data_prob[i] = np.mean(prob)
        data_label_imu[i] = int(np.mean(label_imu) >= 0.5)

    # Set probability to zero if majority IMU label is 0
    data_prob[data_label_imu == 0] = 0
    sqa_label = data_prob >= config.threshold_sqa

    return sqa_label


def extract_pr_segments(sqa_label: np.ndarray, min_pr_samples: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Extracts pulse rate segments based on the SQA label.

    Parameters
    ----------
    sqa_label : np.ndarray
        The signal quality assessment label.
    min_pr_samples : int
        The minimum number of samples required for a pulse rate segment.

    Returns
    -------
    Tuple[v_start_idx_long, v_end_idx_long]
        The start and end indices of the pulse rate segments.
    """
    # Find the start and end indices of the pulse rate segments
    v_start_idx = np.where(np.diff(sqa_label.astype(int)) == 1)[0] + 1
    v_end_idx = np.where(np.diff(sqa_label.astype(int)) == -1)[0] + 1

    # Check if the first segment is a start or end
    if sqa_label[0] == 1:
        v_start_idx = np.insert(v_start_idx, 0, 0)
    if sqa_label[-1] == 1:
        v_end_idx = np.append(v_end_idx, len(sqa_label))

    # Check if the segments are long enough
    v_start_idx_long = v_start_idx[(v_end_idx - v_start_idx) >= min_pr_samples]
    v_end_idx_long = v_end_idx[(v_end_idx - v_start_idx) >= min_pr_samples]

    return v_start_idx_long, v_end_idx_long


def extract_pr_from_segment(
        ppg: np.ndarray, 
        tfd_length: int, 
        fs: int, 
        kern_type: str, 
        kern_params: dict
    ) -> np.ndarray:
    """
    Extracts pulse rate from the time-frequency distribution of the PPG signal.

    Parameters
    ----------
    ppg : np.ndarray
        The preprocessed PPG segment with 2 seconds of padding on both sides to reduce boundary effects.
    tfd_length : int
        Length of each segment (in seconds) to calculate the time-frequency distribution.
    fs : int
        The sampling frequency of the PPG signal.
    kern_type : str
        Type of TFD kernel to use (e.g., 'wvd' for Wigner-Ville distribution).
    kern_params : dict
        Parameters for the specified kernel. Not required for 'wvd', but relevant for other 
        kernels like 'spwvd' or 'swvd'. Default is None.

    Returns
    -------
    np.ndarray
        The estimated pulse rate.
    """

    # Constants to handle boundary effects
    edge_padding = 4 * fs  # Additional 4 seconds (2 seconds on both sides)
    tfd_length = tfd_length * fs  # Convert tfd_length to samples

    # Calculate the actual segment length without padding
    original_segment_length = len(ppg) - edge_padding

    # Determine the number of tfd_length-second segments
    num_segments = max(1, int(original_segment_length // tfd_length))

    # Split the PPG signal into segments
    ppg_segments = []
    for i in range(num_segments):
        if i != num_segments - 1:  # For all segments except the last
            start_idx = int(i * tfd_length)
            end_idx = int((i + 1) * tfd_length + edge_padding)
        else:  # For the last segment
            start_idx = int(i * tfd_length)
            end_idx = len(ppg)
        ppg_segments.append(ppg[start_idx:end_idx])

    pr_est_from_ppg = np.array([])
    for segment in ppg_segments:
        # Calculate the time-frequency distribution
        pr_tfd = extract_pr_with_tfd(segment, fs, kern_type, kern_params)
        pr_est_from_ppg = np.concatenate((pr_est_from_ppg, pr_tfd)) 

    return pr_est_from_ppg


def extract_pr_with_tfd(
        ppg: np.ndarray, 
        fs: int, 
        kern_type: str, 
        kern_params: dict
    ) -> np.ndarray:
    """
    Estimate pulse rate (PR) from a PPG segment using a TFD method with optional 
    moving average filtering.

    Parameters
    ----------
    ppg_segment : array-like
        Segment of the PPG signal to process.
    fs : int
        Sampling frequency in Hz.
    kern_type : str
        Type of TFD kernel to use.
    kern_params : dict
        Parameters for the specified kernel.

    Returns
    -------
    pr_smooth_tfd : np.ndarray
        Estimated pr values (in beats per minute) for each 2-second segment of the PPG signal.
    """
    # Generate the TFD matrix using the specified kernel
    tfd_obj = TimeFreqDistr()
    tfd = tfd_obj.nonsep_gdtfd(ppg, kern_type, kern_params)  # Returns an NxN matrix

    # Get time and frequency axes for the TFD
    num_time_samples, num_freq_bins = tfd.shape
    time_axis = np.arange(num_time_samples) / fs
    freq_axis = np.linspace(0, 0.5, num_freq_bins) * fs

    # Estimate pulse rate by identifying the max frequency in the TFD
    max_freq_indices = np.argmax(tfd, axis=0)

    pr_smooth_tfd = np.array([])
    for i in range(2, int(len(ppg) / fs) - 4 + 1, 2):  # Skip the first and last 2 seconds, add 1 to include the last segment
        relevant_indices = (time_axis >= i) & (time_axis < i + 2)
        avg_frequency = np.mean(freq_axis[max_freq_indices[relevant_indices]])
        pr_smooth_tfd = np.concatenate((pr_smooth_tfd, [60 * avg_frequency]))  # Convert frequency to BPM

    return pr_smooth_tfd


class TimeFreqDistr:
    def __init__(self):
        """
        This module contains the implementation of the Generalized Time-Frequency Distribution (TFD) computation using non-separable kernels.
        This is a Python implementation of the MATLAB code provided by John O Toole in the following repository: https://github.com/otoolej/memeff_TFDs 

        The following functions are implemented for the computation of the TFD:
            - nonsep_gdtfd: Computes the generalized time-frequency distribution using a non-separable kernel.
            - get_analytic_signal: Generates the analytic signal of the input signal.
            - gen_analytic: Generates the analytic signal by zero-padding and performing FFT.
            - gen_time_lag: Generates the time-lag distribution of the analytic signal.
            - multiply_kernel_signal: Multiplies the TFD by the Doppler-lag kernel.
            - gen_doppler_lag_kern: Generates the Doppler-lag kernel based on kernel type and parameters.
            - get_kern: Gets the kernel based on the provided kernel type.
            - get_window: General function to calculate a window function.
            - get_win: Helper function to create the specified window type.
            - shift_window: Shifts the window so that positive indices appear first.
            - pad_window: Zero-pads the window to a specified length.
            - compute_tfd: Finalizes the time-frequency distribution computation.
        """	
        pass

    def nonsep_gdtfd(
            self,
            x: np.ndarray, 
            kern_type: None | str = None,
            kern_params: None | dict = None
        ):
        """
        Computes the generalized time-frequency distribution (TFD) using a non-separable kernel.

        Parameters:
        -----------
        x : ndarray
            Input signal to be analyzed.
        kern_type : str, optional
            Type of kernel to be used for TFD computation. Default is None.
            Currently supported kernels are:
                wvd  - kernel for Wigner-Ville distribution
                swvd - kernel for smoothed Wigner-Ville distribution
                    (lag-independent kernel)
                pwvd - kernel for pseudo Wigner-Ville distribution
                    (Doppler-independent kernel)
                sep  - kernel for separable kernel (combintation of SWVD and PWVD)

        kern_params : dict, optional
            Dictionary of parameters specific to the kernel type. Default is None.
            The structure of the dictionary depends on the selected kernel type:
                - wvd:
                An empty dictionary, as no additional parameters are required.
                - swvd:
                Dictionary with the following keys:
                    'win_length': Length of the smoothing window.
                    'win_type': Type of window function (e.g., 'hamm', 'hann').
                    'win_param' (optional): Additional parameters for the window.
                    'win_param2' (optional): 0 for time-domain window or 1 for Doppler-domain window.

            Example:
            ```python
            kern_params = {
                'win_length': 11,
                'win_type': 'hamm',
                'win_param': 0,
                'domain': 1
            }
            ```
            - pwvd:
            Dictionary with the following keys:
                'win_length': Length of the smoothing window.
                'win_type': Type of window function (e.g., 'cosh').
                'win_param' (optional): Additional parameters for the window.
                'win_param2' (optional): 0 for time-domain window or 1 for Doppler-domain window.
            Example:
            ```python
            kern_params = {
                'win_length': 200,
                'win_type': 'cosh',
                'win_param': 0.1
            }
            ```
            - sep:
            Dictionary containing two nested dictionaries, one for the Doppler window and one for the lag window:
                'doppler': {
                    'win_length': Length of the Doppler-domain window.
                    'win_type': Type of Doppler-domain window function.
                    'win_param' (optional): Additional parameters for the Doppler window.
                    'win_param2' (optional): 0 for time-domain window or 1 for Doppler-domain window.
                }
                'lag': {
                    'win_length': Length of the lag-domain window.
                    'win_type': Type of lag-domain window function.
                    'win_param' (optional): Additional parameters for the lag window.
                    'win_param2' (optional): 0 for time-domain window or 1 for Doppler-domain window.
                }
            Example:
            ```python
            kern_params = {
                'doppler': {
                    'win_length': doppler_samples,
                    'win_type': win_type_doppler,
                },
                'lag': {
                    'win_length': lag_samples,
                    'win_type': win_type_lag,
                }
            }
            ```

        Returns:
        --------
        tfd : ndarray
            The computed time-frequency distribution.
        """
        z = self.get_analytic_signal(x)
        N = len(z) // 2  # Since z is a signal of length 2N
        Nh = int(np.ceil(N / 2))

        # Generate the time-lag distribution of the analytic signal
        tfd = self.gen_time_lag(z)

        # Multiply the TFD by the Doppler-lag kernel
        tfd = self.multiply_kernel_signal(tfd, kern_type, kern_params, N, Nh)
        
        # Finalize the TFD computation
        tfd = self.compute_tfd(N, Nh, tfd)
        
        return tfd

    def get_analytic_signal(self, x: np.ndarray) -> np.ndarray:
        """
        Generates the signals analytic version.

        Parameters:
        -----------
        z : ndarray
            Input real-valued signal.

        Returns:
        --------
        z : ndarray
            Analytic signal with zero-padded imaginary part.
        """
        N = len(x)

        # Ensure the signal length is even by trimming one sample if odd, since the gen_time_lag function requires an even-length signal
        if N % 2 != 0:
            x = x[:-1]

        # Make the analytical signal of the real-valued signal z (preprocessed PPG signal)
        # doesn't work for input of complex numbers
        z = self.gen_analytic(x)  
        
        return z

    def gen_analytic(self, x: np.ndarray) -> np.ndarray:
        """
        Generates an analytic signal by zero-padding and performing FFT.

        Parameters:
        -----------
        x : ndarray
            Input real-valued signal.

        Returns:
        --------
        z : ndarray
            Analytic signal in the time domain with zeroed second half.
        """
        N = len(x)
        
        # Zero-pad the signal to double its length
        x = np.concatenate((np.real(x), np.zeros(N)))
        x_fft = np.fft.fft(x)

        # Generate the analytic signal in the frequency domain
        H = np.empty(2 * N) # Preallocate an array of size 2*N
        H[0] = 1 # First element
        H[1:N] = 2 # Next N-1 elements
        H[N] = 1 # Middle element
        H[N+1:] = 0 # Last N-1 elements
        z_cb = np.fft.ifft(x_fft * H)

        # Force the second half of the time-domain signal to zero
        z = np.concatenate((z_cb[:N], np.zeros(N)))

        return z

    def gen_time_lag(self, z: np.ndarray) -> np.ndarray:
        """
        Generate the time-lag distribution of the analytic signal z.

        Parameters:
        -----------
        z : ndarray
            Analytic signal of the input signal x.
        
        Returns:
        --------
        tfd : ndarray
            Time-lag distribution of the analytic signal z.

        """
        N = len(z) // 2  # Assuming z is a signal of length 2N
        Nh = int(np.ceil(N / 2))

        # Initialize the time-frequency distribution (TFD) matrix
        tfd = np.zeros((N, N), dtype=complex)

        m = np.arange(Nh)
        
        # Loop over time indices
        for n in range(N):
            inp = np.mod(n + m, 2 * N)
            inn = np.mod(n - m, 2 * N)

            # Extract the time slice from the analytic signal
            K_time_slice = z[inp] * np.conj(z[inn])

            # Store real and imaginary parts
            tfd[n, :Nh] = np.real(K_time_slice)
            tfd[n, Nh:] = np.imag(K_time_slice)
        
        return tfd

    def multiply_kernel_signal(  
        self,
        tfd: np.ndarray, 
        kern_type: str, 
        kern_params: dict, 
        N: int, 
        Nh: int  
        ) -> np.ndarray:  
        """
        Multiplies the TFD by the Doppler-lag kernel.

        Parameters:
        -----------
        tfd : ndarray
            Time-frequency distribution.
        kern_type : str
            Kernel type to be applied.
        kern_params : dict
            Kernel parameters specific to the kernel type.
        N : int
            Length of the signal.
        Nh : int
            Half length of the signal.

        Returns:
        --------
        tfd : ndarray
            Modified TFD after kernel multiplication.
        """
        # Loop over lag indices
        for m in range(Nh):
            # Generate the Doppler-lag kernel for each lag index
            g_lag_slice = self.gen_doppler_lag_kern(kern_type, kern_params, N, m)
            
            # Extract and transform the TFD slice for this lag
            tfd_slice = np.fft.fft(tfd[:, m]) + 1j * np.fft.fft(tfd[:, Nh + m])
            
            # Multiply by the kernel and perform inverse FFT
            R_lag_slice = np.fft.ifft(tfd_slice * g_lag_slice)
            
            # Store real and imaginary parts back into the TFD
            tfd[:, m] = np.real(R_lag_slice)
            tfd[:, Nh + m] = np.imag(R_lag_slice)
        
        return tfd

    def gen_doppler_lag_kern(
            self, 
            kern_type: str, 
            kern_params: dict, 
            N: int, 
            lag_index: int
        ):
        """
        Generate the Doppler-lag kernel based on kernel type and parameters.

        Parameters:
        -----------
        kern_type : str
            Type of kernel (e.g., 'wvd', 'swvd', 'pwvd', etc.).
        kern_params : dict
            Parameters for the kernel.
        N : int
            Signal length.
        lag_index : int
            Current lag index.

        Returns:
        --------
        g : ndarray
            Doppler-lag kernel for the given lag.
        """
        g = np.zeros(N, dtype=complex)  # Initialize the kernel

        # Get kernel based on the type
        g = self.get_kern(g, lag_index, kern_type, kern_params, N)

        return np.real(g) # All kernels are real valued

    def get_kern(
            self, 
            g: np.ndarray, 
            lag_index: int, 
            kern_type: str, 
            kern_params: dict, 
            N: int
        ) -> np.ndarray:
        """
        Get the kernel based on the provided kernel type.

        Parameters:
        -----------
        g : ndarray
            Kernel to be filled.
        lag_index : int
            Lag index for the kernel.
        kern_type : str
            Type of kernel to use (now included: 'wvd', 'swvd', 'pwvd', 'sep').
        kern_params : dict
            Parameters for the specified kernel.
        N : int
            Signal length.

        Returns:
        --------
        g : ndarray
            Kernel function at the current lag.
        """
        # Validate kern_type
        valid_kern_types = ['wvd', 'sep', 'swvd', 'pwvd']  # List of valid kernel types which are currently supported
        if kern_type not in valid_kern_types:
            raise ValueError(f"Unknown kernel type: {kern_type}. Expected one of {valid_kern_types}")
        
        num_params = len(kern_params)

        if kern_type == 'wvd':
            g[:] = 1 # WVD kernel is the equal to 1 for all lags

        elif kern_type == 'sep':
            # Separable Kernel
            g1 = np.copy(g)  # Create a new array for g1
            g2 = np.copy(g)  # Create a new array for g2
            
            # Call recursively to obtain g1 and g2 kernels (no in-place modification of g)
            g1 = self.get_kern(g1, lag_index, 'swvd', kern_params['lag'], N) # Generate the first kernel
            g2 = self.get_kern(g2, lag_index, 'pwvd', kern_params['doppler'], N) # Generate the second kernel
            g = g1 * g2 # Multiply the two kernels to obtain the separable kernel

        else:
            if num_params < 2:
                raise ValueError("Missing required kernel parameters: 'win_length' and 'win_type'")

            win_length = kern_params['win_length']
            win_type = kern_params['win_type']
            win_param = kern_params['win_param'] if 'win_param' in kern_params else 0
            win_param2 = kern_params['win_param2'] if 'win_param2' in kern_params else 1

            G = self.get_window(win_length, win_type, win_param)
            G = self.pad_window(G, N)

            if kern_type == 'swvd' and win_param2 == 0:
                G = np.fft.fft(G)
                if G[0] != 0:  # add this check to avoid division by zero
                    G /= G[0]
                G = G[lag_index]

            g[:] = G

        return g

    def get_window(
            self, 
            win_length: int, 
            win_type: str,
            win_param: float | None = None,
            dft_window: bool = False,
            Npad: int = 0
        ) -> np.ndarray:
        """
        General function to calculate a window function.
        
        Parameters:
        -----------
        win_length : int
            Length of the window.
        win_type : str
            Type of window. Options are:
            {'delta', 'rect', 'hamm'|'hamming', 'hann'|'hanning', 'gauss', 'cosh'}.
        win_param : float, optional
            Window parameter (e.g., alpha for Gaussian window). Default is None.
        dft_window : bool, optional
            If True, returns the DFT of the window. Default is False.
        Npad : int, optional
            If greater than 0, zero-pads the window to length Npad. Default is 0.
        
        Returns:
        --------
        win : ndarray
            The calculated window (or its DFT if dft_window is True).
        """
        
        # Get the window
        win = self.get_win(win_length, win_type, win_param, dft_window)
        
        # Shift the window so that positive indices are first
        win = self.shift_window(win)
        
        # Zero-pad the window to length Npad if necessary
        if Npad > 0:
            win = self.pad_window(win, Npad)
        
        return win

    def get_win(
            self, 
            win_length: int, 
            win_type: str, 
            win_param: float | None = None, 
            dft_window: bool = False
        ) -> np.ndarray:
        """
        Helper function to create the specified window type.
        
        Parameters:
        -----------
        win_length : int
            Length of the window.
        win_type : str
            Type of window.
        win_param : float, optional
            Additional parameter for certain window types (e.g., Gaussian alpha). Default is None.
        dft_window : bool, optional
            If True, returns the DFT of the window. Default is False.
        
        Returns:
        --------
        win : ndarray
            The created window (or its DFT if dft_window is True).
        """
        if win_type == 'delta':
            win = np.zeros(win_length)
            win[win_length // 2] = 1
        elif win_type == 'rect':
            win = np.ones(win_length)
        elif win_type in ['hamm', 'hamming']:
            win = signal.windows.hamming(win_length)
        elif win_type in ['hann', 'hanning']:
            win = signal.windows.hann(win_length)
        elif win_type == 'gauss':
            win = signal.windows.gaussian(win_length, std=win_param if win_param else 0.4)
        elif win_type == 'cosh':
            win_hlf = win_length // 2
            if not win_param:
                win_param = 0.01
            win = np.array([np.cosh(m) ** (-2 * win_param) for m in range(-win_hlf, win_hlf+1)])
            win = np.fft.fftshift(win)
        else:
            raise ValueError(f"Unknown window type {win_type}")
        
        # If dft_window is True, return the DFT of the window
        if dft_window:
            win = np.fft.fft(np.roll(win, win_length // 2))
            win = np.roll(win, -win_length // 2)
        
        return win

    def shift_window(self, w: np.ndarray) -> np.ndarray:
        """
        Shift the window so that positive indices appear first.
        
        Parameters:
        -----------
        w : ndarray
            Window to be shifted.
        
        Returns:
        --------
        w_shifted : ndarray
            Shifted window with positive indices first.
        """
        N = len(w)
        return np.roll(w, N // 2)

    def pad_window(self, w: np.ndarray, Npad: int) -> np.ndarray:
        """
        Zero-pad the window to a specified length.
        
        Parameters:
        -----------
        w : ndarray
            The original window.
        Npad : int
            Length to zero-pad the window to.
        
        Returns:
        --------
        w_pad : ndarray
            Zero-padded window of length Npad.
        
        Raises:
        -------
        ValueError:
            If Npad is less than the original window length.
        """
        N = len(w)
        w_pad = np.zeros(Npad)
        Nh = N // 2
        
        if Npad < N:
            raise ValueError("Npad must be greater than or equal to the window length")

        if N == Npad:
            return w
        
        if N % 2 == 1:  # For odd N
            w_pad[:Nh+1] = w[:Nh+1]
            w_pad[-Nh:] = w[-Nh:]
        else:  # For even N
            w_pad[:Nh] = w[:Nh]
            w_pad[Nh] = w[Nh] / 2
            w_pad[-Nh:] = w[-Nh:]
            w_pad[-Nh] = w[Nh] / 2
        
        return w_pad

    def compute_tfd(
            self, 
            N: int, 
            Nh: int, 
            tfd: np.ndarray
        ):
        """
        Finalizes the time-frequency distribution computation.

        Parameters:
        -----------
        N : int
            Size of the TFD.
        Nh : int
            Half-length parameter.
        tfd : np.ndarray
            Time-frequency distribution to be finalized.

        Returns:
        --------
        tfd : np.ndarray
            Final computed TFD (N,N).
        """
        m = np.arange(0, Nh)  # m = 0:(Nh-1)
        mb = np.arange(1, Nh)  # mb = 1:(Nh-1)

        for n in range(0, N-1, 2):  # for n=0:2:N-2
            R_even_half = np.complex128(tfd[n, :Nh]) + 1j * np.complex128(tfd[n, Nh:])
            R_odd_half = np.complex128(tfd[n+1, :Nh]) + 1j * np.complex128(tfd[n+1, Nh:])

            R_tslice_even = np.zeros(N, dtype=np.complex128)  
            R_tslice_odd = np.zeros(N, dtype=np.complex128)

            R_tslice_even[m] = R_even_half
            R_tslice_odd[m] = R_odd_half

            R_tslice_even[N-mb] = np.conj(R_even_half[mb])  
            R_tslice_odd[N-mb] = np.conj(R_odd_half[mb])
            
            # Perform FFT to compute time slices
            tfd_time_slice = np.fft.fft(R_tslice_even + 1j * R_tslice_odd)

            tfd[n, :] = np.real(tfd_time_slice)
            tfd[n+1, :] = np.imag(tfd_time_slice)

        tfd = tfd / N  # Normalize the TFD
        tfd = tfd.transpose()  # Transpose the TFD to have the time on the x-axis and frequency on the y-axis   
        return tfd



from typing import Dict, List
from paradigma.constants import DataColumns, DataUnits
import numpy as np

class BaseConfig:
    def __init__(self) -> None:
        self.meta_filename = ''
        self.values_filename = ''
        self.time_filename = ''

    def set_sensor(self, sensor: str) -> None:
        """Sets the sensor and derived filenames"""
        self.sensor: str = sensor
        self.set_filenames(sensor)

    def set_filenames(self, prefix: str) -> None:
        """Sets the filenames based on the prefix. This method is duplicated from `gaits_analysis_config.py`.
        
        Parameters
        ----------
        prefix : str
            The prefix for the filenames.
        """
        self.meta_filename = f"{prefix}_meta.json"
        self.time_filename = f"{prefix}_time.bin"
        self.values_filename = f"{prefix}_values.bin"

class IMUConfig(BaseConfig):

    def __init__(self) -> None:
        super().__init__()

        self.set_filenames('IMU')

        self.acceleration_units = DataUnits.ACCELERATION
        self.rotation_units = DataUnits.ROTATION

        self.axes = ["x", "y", "z"]

        self.accelerometer_cols: List[str] = [
            DataColumns.ACCELEROMETER_X,
            DataColumns.ACCELEROMETER_Y,
            DataColumns.ACCELEROMETER_Z,
        ]
        self.gyroscope_cols: List[str] = [
            DataColumns.GYROSCOPE_X,
            DataColumns.GYROSCOPE_Y,
            DataColumns.GYROSCOPE_Z,
        ]
        self.gravity_cols: List[str] = [
            DataColumns.GRAV_ACCELEROMETER_X,
            DataColumns.GRAV_ACCELEROMETER_Y,
            DataColumns.GRAV_ACCELEROMETER_Z,
        ]

        self.d_channels_accelerometer = {
            DataColumns.ACCELEROMETER_X: self.acceleration_units,
            DataColumns.ACCELEROMETER_Y: self.acceleration_units,
            DataColumns.ACCELEROMETER_Z: self.acceleration_units,
        }
        self.d_channels_gyroscope = {
            DataColumns.GYROSCOPE_X: self.rotation_units,
            DataColumns.GYROSCOPE_Y: self.rotation_units,
            DataColumns.GYROSCOPE_Z: self.rotation_units,
        }
        self.d_channels_imu = {**self.d_channels_accelerometer, **self.d_channels_gyroscope}

        self.sampling_frequency = 100
        self.lower_cutoff_frequency = 0.2
        self.upper_cutoff_frequency = 3.5
        self.filter_order = 4


class PPGConfig(BaseConfig):

    def __init__(self) -> None:
        super().__init__()

        self.set_filenames('PPG')

        self.ppg_colname = DataColumns.PPG

        self.sampling_frequency = 30
        self.lower_cutoff_frequency = 0.4
        self.upper_cutoff_frequency = 3.5
        self.filter_order = 4

        self.d_channels_ppg = {
            DataColumns.PPG: DataUnits.NONE
        }


# Domain base configs
class GaitConfig(IMUConfig):

    def __init__(self, step) -> None:
        super().__init__()

        self.set_sensor('accelerometer')

        # ----------
        # Segmenting
        # ----------
        self.max_segment_gap_s = 1.5
        self.min_segment_length_s = 1.5

        if step == 'gait':
            self.window_length_s: float = 6
            self.window_step_length_s: float = 1
        else:
            self.window_length_s: float = 3
            self.window_step_length_s: float = self.window_length_s * 0.25

        # -----------------
        # Feature extraction
        # -----------------
        self.window_type: str = "hann"
        self.spectrum_low_frequency: int = 0
        self.spectrum_high_frequency: int = int(self.sampling_frequency / 2)

        # Power in specified frequency bands
        self.d_frequency_bandwidths: Dict[str, List[float]] = {
            "power_below_gait": [0.2, 0.7],
            "power_gait": [0.7, 3.5],
            "power_tremor": [3.5, 8],
            "power_above_tremor": [8, 25],
        }

        # Mel frequency cepstral coefficients
        self.mfcc_low_frequency: int = 0
        self.mfcc_high_frequency: int = 25
        self.mfcc_n_dct_filters: int = 15
        self.mfcc_n_coefficients: int = 12

        # Dominant frequency of first harmonic of arm swing
        self.angle_fmin: float = 0.5
        self.angle_fmax: float = 1.5

        # -----------------
        # TSDF data storage
        # -----------------
        self.d_channels_values: Dict[str, str] = {
            "accelerometer_std_norm": DataUnits.GRAVITY,
            "accelerometer_x_grav_mean": DataUnits.GRAVITY,
            "accelerometer_y_grav_mean": DataUnits.GRAVITY,
            "accelerometer_z_grav_mean": DataUnits.GRAVITY,
            "accelerometer_x_grav_std": DataUnits.GRAVITY,
            "accelerometer_y_grav_std": DataUnits.GRAVITY,
            "accelerometer_z_grav_std": DataUnits.GRAVITY,
            "accelerometer_x_power_below_gait": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_y_power_below_gait": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_z_power_below_gait": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_x_power_gait": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_y_power_gait": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_z_power_gait": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_x_power_tremor": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_y_power_tremor": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_z_power_tremor": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_x_power_above_tremor": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_y_power_above_tremor": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_z_power_above_tremor": DataUnits.POWER_SPECTRAL_DENSITY,
            "accelerometer_x_dominant_frequency": DataUnits.FREQUENCY,
            "accelerometer_y_dominant_frequency": DataUnits.FREQUENCY,
            "accelerometer_z_dominant_frequency": DataUnits.FREQUENCY,
        }

        for mfcc_coef in range(1, self.mfcc_n_coefficients + 1):
            self.d_channels_values[f"accelerometer_mfcc_{mfcc_coef}"] = DataUnits.GRAVITY

        if step == 'arm_activity':
            for mfcc_coef in range(1, self.mfcc_n_coefficients + 1):
                self.d_channels_values[f"gyroscope_mfcc_{mfcc_coef}"] = DataUnits.GRAVITY


class TremorConfig(IMUConfig):

    def __init__(self, step: str | None = None) -> None:
        """
        Parameters
        ----------
        step : str (optional)
            The step of the tremor pipeline. Can be 'features' or 'classification'.
        """
        super().__init__()

        self.set_sensor('gyroscope')

        # ----------
        # Segmenting
        # ----------
        self.window_length_s: float = 4
        self.window_step_length_s: float = 4

        # -----------------
        # Feature extraction
        # -----------------
        self.window_type = 'hann'
        self.overlap_fraction: float = 0.8
        self.segment_length_psd_s: float = 3
        self.segment_length_spectrogram_s: float = 2
        self.spectral_resolution: float = 0.25
        
        # PSD based features
        self.fmin_peak_search: float = 1
        self.fmax_peak_search: float = 25
        self.fmin_below_rest_tremor: float = 0.5
        self.fmax_below_rest_tremor: float = 3
        self.fmin_rest_tremor: float = 3
        self.fmax_rest_tremor: float = 7

        # Mel frequency cepstral coefficients
        self.fmin_mfcc: float = 0
        self.fmax_mfcc: float = 25
        self.n_dct_filters_mfcc: int = 15
        self.n_coefficients_mfcc: int = 12

        # --------------
        # Classification
        # --------------
        self.movement_threshold: float = 50

        # -----------
        # Aggregation
        # -----------
        self.aggregates_tremor_power: List[str] = ['mode', 'median', '90p']

        # -----------------
        # TSDF data storage
        # -----------------
        if step == 'features':
            self.d_channels_values: Dict[str, str] = {}
            for mfcc_coef in range(1, self.n_coefficients_mfcc + 1):
                self.d_channels_values[f"mfcc_{mfcc_coef}"] = "unitless"

            self.d_channels_values["freq_peak"] = "Hz"
            self.d_channels_values["below_tremor_power"] = "(deg/s)^2"
            self.d_channels_values["tremor_power"] = "(deg/s)^2"
        elif step == 'classification':
            self.d_channels_values = {
                DataColumns.PRED_TREMOR_PROBA: "probability",
                DataColumns.PRED_TREMOR_LOGREG: "boolean",
                DataColumns.PRED_TREMOR_CHECKED: "boolean",
                DataColumns.PRED_ARM_AT_REST: "boolean"
            }

        
class PulseRateConfig(PPGConfig):
    def __init__(self, sensor: str = 'ppg', min_window_length_s: int = 30) -> None:
        super().__init__()

        # ----------
        # Segmenting
        # ----------
        self.window_length_s: int = 6
        self.window_step_length_s: int = 1
        self.window_overlap_s = self.window_length_s - self.window_step_length_s

        self.accelerometer_cols = IMUConfig().accelerometer_cols

        # -----------------------
        # Signal quality analysis
        # -----------------------
        self.freq_band_physio = [0.75, 3] # Hz
        self.bandwidth = 0.2   # Hz      
        self.freq_bin_resolution = 0.05 # Hz

        # ---------------------
        # Pulse rate estimation
        # ---------------------
        self.set_tfd_length(min_window_length_s)  # Set tfd length to default of 30 seconds
        self.threshold_sqa = 0.5
        self.threshold_sqa_accelerometer = 0.10

        pr_est_length = 2  # pulse rate estimation length in seconds
        self.pr_est_samples = pr_est_length * self.sampling_frequency

        # Time-frequency distribution parameters
        self.kern_type = 'sep'
        win_type_doppler = 'hamm'
        win_type_lag = 'hamm'
        win_length_doppler = 8
        win_length_lag = 1
        doppler_samples = self.sampling_frequency * win_length_doppler
        lag_samples = win_length_lag * self.sampling_frequency
        self.kern_params = {
            'doppler': {
                'win_length': doppler_samples,
                'win_type': win_type_doppler,
            },
            'lag': {
                'win_length': lag_samples,
                'win_type': win_type_lag,
            }
        }

        self.set_sensor(sensor)

    def set_tfd_length(self, tfd_length: int):
        self.tfd_length = tfd_length
        self.min_pr_samples = int(round(self.tfd_length * self.sampling_frequency))

    def set_sensor(self, sensor):
        self.sensor = sensor 

        if sensor not in ['ppg', 'imu']:
            raise ValueError(f"Invalid sensor type: {sensor}")
        
        if sensor == 'imu':
            self.sampling_frequency = IMUConfig().sampling_frequency
        else:
            self.sampling_frequency = PPGConfig().sampling_frequency

        self.window_length_welch = 3 * self.sampling_frequency
        self.overlap_welch_window = self.window_length_welch // 2
        self.nfft = len(np.arange(0, self.sampling_frequency/2, self.freq_bin_resolution))*2



import json
import numpy as np
import os
import pandas as pd
from pathlib import Path
from scipy.signal import welch
from scipy.signal.windows import hamming, hann
import tsdf
from typing import List

from paradigma.classification import ClassifierPackage
from paradigma.constants import DataColumns
from paradigma.config import PulseRateConfig
from paradigma.feature_extraction import compute_statistics, compute_signal_to_noise_ratio, compute_auto_correlation, \
    compute_dominant_frequency, compute_relative_power, compute_spectral_entropy
from paradigma.pipelines.pulse_rate_utils import assign_sqa_label, extract_pr_segments, extract_pr_from_segment
from paradigma.segmenting import tabulate_windows, WindowedDataExtractor
from paradigma.util import aggregate_parameter

def extract_signal_quality_features(df_ppg: pd.DataFrame, df_acc: pd.DataFrame, ppg_config: PulseRateConfig, acc_config: PulseRateConfig) -> pd.DataFrame:
    """	
    Extract signal quality features from the PPG signal.
    The features are extracted from the temporal and spectral domain of the PPG signal.
    The temporal domain features include variance, mean, median, kurtosis, skewness, signal-to-noise ratio, and autocorrelation.
    The spectral domain features include the dominant frequency, relative power, spectral entropy.

    Parameters
    ----------
    df_ppg : pd.DataFrame
        The DataFrame containing the PPG signal.
    df_acc : pd.DataFrame
        The DataFrame containing the accelerometer signal.
    ppg_config: PulseRateConfig
        The configuration for the signal quality feature extraction of the PPG signal.
    acc_config: PulseRateConfig
        The configuration for the signal quality feature extraction of the accelerometer signal.

    Returns
    -------
    df_features : pd.DataFrame
        The DataFrame containing the extracted signal quality features.
    
    """
    # Group sequences of timestamps into windows
    ppg_windowed_cols = [DataColumns.TIME, ppg_config.ppg_colname]
    ppg_windowed = tabulate_windows(
        df=df_ppg, 
        columns=ppg_windowed_cols,
        window_length_s=ppg_config.window_length_s,
        window_step_length_s=ppg_config.window_step_length_s,
        fs=ppg_config.sampling_frequency
    )

    # Extract data from the windowed PPG signal
    extractor = WindowedDataExtractor(ppg_windowed_cols)
    idx_time = extractor.get_index(DataColumns.TIME)
    idx_ppg = extractor.get_index(ppg_config.ppg_colname)
    start_time_ppg = np.min(ppg_windowed[:, :, idx_time], axis=1) # Start time of the window is relative to the first datapoint in the PPG data
    ppg_values_windowed = ppg_windowed[:, :, idx_ppg]

    acc_windowed_cols = [DataColumns.TIME] + acc_config.accelerometer_cols
    acc_windowed = tabulate_windows(
        df=df_acc,
        columns=acc_windowed_cols,
        window_length_s=acc_config.window_length_s,
        window_step_length_s=acc_config.window_step_length_s,
        fs=acc_config.sampling_frequency
    )

    # Extract data from the windowed accelerometer signal
    extractor = WindowedDataExtractor(acc_windowed_cols)
    idx_acc = extractor.get_slice(acc_config.accelerometer_cols)
    acc_values_windowed = acc_windowed[:, :, idx_acc]

    df_features = pd.DataFrame(start_time_ppg, columns=[DataColumns.TIME])
    # Compute features of the temporal domain of the PPG signal
    df_temporal_features = extract_temporal_domain_features(ppg_values_windowed, ppg_config, quality_stats=['var', 'mean', 'median', 'kurtosis', 'skewness'])
    
    # Combine temporal features with the start time
    df_features = pd.concat([df_features, df_temporal_features], axis=1)

    # Compute features of the spectral domain of the PPG signal
    df_spectral_features = extract_spectral_domain_features(ppg_values_windowed, ppg_config)

    # Combine the spectral features with the previously computed temporal features
    df_features = pd.concat([df_features, df_spectral_features], axis=1)
    
    # Compute periodicity feature of the accelerometer signal
    df_accelerometer_feature = extract_accelerometer_feature(acc_values_windowed, ppg_values_windowed, acc_config)

    # Combine the accelerometer feature with the previously computed features
    df_features = pd.concat([df_features, df_accelerometer_feature], axis=1)

    return df_features


def signal_quality_classification(df: pd.DataFrame, config: PulseRateConfig, full_path_to_classifier_package: str | Path) -> pd.DataFrame:
    """
    Classify the signal quality of the PPG signal using a logistic regression classifier. A probability close to 1 indicates a high-quality signal, while a probability close to 0 indicates a low-quality signal.
    The classifier is trained on features extracted from the PPG signal. The features are extracted using the extract_signal_quality_features function.
    The accelerometer signal is used to determine the signal quality based on the power ratio of the accelerometer signal and returns a binary label based on a threshold.
    A value of 1 on the indicates no/minor periodic motion influence of the accelerometer on the PPG signal, 0 indicates major periodic motion influence.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing the PPG features and the accelerometer feature for signal quality classification.
    config : PulseRateConfig
        The configuration for the signal quality classification.
    full_path_to_classifier_package : str | Path
        The path to the directory containing the classifier.

    Returns
    -------
    df_sqa pd.DataFrame
        The DataFrame containing the PPG signal quality predictions (both probabilities of the PPG signal quality classification and the accelerometer label based on the threshold).
    """
    clf_package = ClassifierPackage.load(full_path_to_classifier_package)  # Load the classifier package
    clf = clf_package.classifier  # Load the logistic regression classifier

    # Apply scaling to relevant columns
    scaled_features = clf_package.transform_features(df.loc[:, clf.feature_names_in]) # Apply scaling to the features 

    # Make predictions for PPG signal quality assessment, and assign the probabilities to the DataFrame and drop the features
    df[DataColumns.PRED_SQA_PROBA] = clf.predict_proba(scaled_features)[:, 0]
    df[DataColumns.PRED_SQA_ACC_LABEL] = (df[DataColumns.ACC_POWER_RATIO] < config.threshold_sqa_accelerometer).astype(int)  # Assign accelerometer label to the DataFrame based on the threshold
    
    return df[[DataColumns.TIME, DataColumns.PRED_SQA_PROBA, DataColumns.PRED_SQA_ACC_LABEL]]  # Return only the relevant columns, namely the predicted probabilities for the PPG signal quality and the accelerometer label


def estimate_pulse_rate(df_sqa: pd.DataFrame, df_ppg_preprocessed: pd.DataFrame, config: PulseRateConfig) -> pd.DataFrame:  
    """
    Estimate the pulse rate from the PPG signal using the time-frequency domain method.

    Parameters
    ----------
    df_sqa : pd.DataFrame
        The DataFrame containing the signal quality assessment predictions.
    df_ppg_preprocessed : pd.DataFrame
        The DataFrame containing the preprocessed PPG signal.
    config : PulseRateConfig
        The configuration for the pulse rate estimation.

    Returns
    -------
    df_pr : pd.DataFrame
        The DataFrame containing the pulse rate estimations.
    """

    # Extract NumPy arrays for faster operations
    ppg_post_prob = df_sqa[DataColumns.PRED_SQA_PROBA].to_numpy()
    acc_label = df_sqa.loc[:, DataColumns.PRED_SQA_ACC_LABEL].to_numpy() # Adjust later in data columns to get the correct label, should be first intergrated in feature extraction and classification
    ppg_preprocessed = df_ppg_preprocessed.values 
    time_idx = df_ppg_preprocessed.columns.get_loc(DataColumns.TIME) # Get the index of the time column
    ppg_idx = df_ppg_preprocessed.columns.get_loc(DataColumns.PPG) # Get the index of the PPG column
    
    # Assign window-level probabilities to individual samples
    sqa_label = assign_sqa_label(ppg_post_prob, config, acc_label) # assigns a signal quality label to every individual data point
    v_start_idx, v_end_idx = extract_pr_segments(sqa_label, config.min_pr_samples) # extracts pulse rate segments based on the SQA label
    
    v_pr_rel = np.array([])
    t_pr_rel = np.array([])

    edge_add = 2 * config.sampling_frequency  # Add 2s on both sides of the segment for PR estimation
    step_size = config.pr_est_samples  # Step size for PR estimation

    # Estimate the maximum size for preallocation
    valid_segments = (v_start_idx >= edge_add) & (v_end_idx <= len(ppg_preprocessed) - edge_add) # check if the segments are valid, e.g. not too close to the edges (2s)
    valid_start_idx = v_start_idx[valid_segments]   # get the valid start indices
    valid_end_idx = v_end_idx[valid_segments]    # get the valid end indices
    max_size = np.sum((valid_end_idx - valid_start_idx) // step_size) # maximum size for preallocation
  
    # Preallocate arrays
    v_pr_rel = np.empty(max_size, dtype=float) 
    t_pr_rel = np.empty(max_size, dtype=float)  

    # Track current position
    pr_pos = 0

    for start_idx, end_idx in zip(valid_start_idx, valid_end_idx):
        # Extract extended PPG segment
        extended_ppg_segment = ppg_preprocessed[start_idx - edge_add : end_idx + edge_add, ppg_idx]

        # Estimate pulse rate
        pr_est = extract_pr_from_segment(
            extended_ppg_segment,
            config.tfd_length,
            config.sampling_frequency,
            config.kern_type,
            config.kern_params,
        )
        n_pr = len(pr_est)  # Number of pulse rate estimates
        end_idx_time = n_pr * step_size + start_idx  # Calculate end index for time, different from end_idx since it is always a multiple of step_size, while end_idx is not

        # Extract relative time for PR estimates
        pr_time = ppg_preprocessed[start_idx : end_idx_time : step_size, time_idx]

        # Insert into preallocated arrays
        v_pr_rel[pr_pos:pr_pos + n_pr] = pr_est
        t_pr_rel[pr_pos:pr_pos + n_pr] = pr_time
        pr_pos += n_pr

    df_pr = pd.DataFrame({"time": t_pr_rel, "pulse_rate": v_pr_rel})

    return df_pr


def aggregate_pulse_rate(pr_values: np.ndarray, aggregates: List[str] = ['mode', '99p']) -> dict:
    """
    Aggregate the pulse rate estimates using the specified aggregation methods.

    Parameters
    ----------
    pr_values : np.ndarray
        The array containing the pulse rate estimates
    aggregates : List[str]
        The list of aggregation methods to be used for the pulse rate estimates. The default is ['mode', '99p'].

    Returns
    -------
    aggregated_results : dict
        The dictionary containing the aggregated results of the pulse rate estimates.
    """
    # Initialize the dictionary for the aggregated results
    aggregated_results = {}

    # Initialize the dictionary for the aggregated results with the metadata
    aggregated_results = {
    'metadata': {
        'nr_pr_est': len(pr_values)
    },
    'pr_aggregates': {}
}
    for aggregate in aggregates:
        aggregated_results['pr_aggregates'][f'{aggregate}_{DataColumns.PULSE_RATE}'] = aggregate_parameter(pr_values, aggregate)

    return aggregated_results


def extract_temporal_domain_features(
        ppg_windowed: np.ndarray, 
        config: PulseRateConfig, 
        quality_stats: List[str] = ['mean', 'std']
    ) -> pd.DataFrame:
    """
    Compute temporal domain features for the ppg signal. The features are added to the dataframe. Therefore the original dataframe is modified, and the modified dataframe is returned.

    Parameters
    ----------
    ppg_windowed: np.ndarray
        The dataframe containing the windowed accelerometer signal

    config: PulseRateConfig
        The configuration object containing the parameters for the feature extraction

    quality_stats: list, optional
        The statistics to be computed for the gravity component of the accelerometer signal (default: ['mean', 'std'])
    
    Returns
    -------
    pd.DataFrame
        The dataframe with the added temporal domain features.
    """
    
    feature_dict = {}
    for stat in quality_stats:
        feature_dict[stat] = compute_statistics(ppg_windowed, stat, abs_stats=True)
    
    feature_dict['signal_to_noise'] = compute_signal_to_noise_ratio(ppg_windowed)  
    feature_dict['auto_corr'] = compute_auto_correlation(ppg_windowed, config.sampling_frequency)
    return pd.DataFrame(feature_dict)


def extract_spectral_domain_features(
        ppg_windowed: np.ndarray,
        config: PulseRateConfig, 
    ) -> pd.DataFrame:
    """
    Calculate the spectral features (dominant frequency, relative power, and spectral entropy)
    for each segment of a PPG signal using a single Welch's method computation. The features are added to the dataframe. 
    Therefore the original dataframe is modified, and the modified dataframe is returned.

    Parameters
    ----------
    ppg_windowed: np.ndarray
        The dataframe containing the windowed ppg signal

    config: PulseRateConfig
        The configuration object containing the parameters for the feature extraction

    Returns
    -------
    pd.DataFrame
        The dataframe with the added spectral domain features.
    """
    d_features = {}

    window = hamming(config.window_length_welch, sym = True)

    n_samples_window = ppg_windowed.shape[1]

    freqs, psd = welch(
        ppg_windowed,
        fs=config.sampling_frequency,
        window=window,
        noverlap=config.overlap_welch_window,
        nfft=max(256, 2 ** int(np.log2(n_samples_window))),
        detrend=False,
        axis=1
    )

    # Calculate each feature using the computed PSD and frequency array
    d_features['f_dom'] = compute_dominant_frequency(freqs, psd)
    d_features['rel_power'] = compute_relative_power(freqs, psd, config)
    d_features['spectral_entropy'] = compute_spectral_entropy(psd, n_samples_window) 

    return pd.DataFrame(d_features)


def extract_acc_power_feature(
        f1: np.ndarray, 
        PSD_acc: np.ndarray, 
        f2: np.ndarray, 
        PSD_ppg: np.ndarray
    ) -> np.ndarray:
    """
    Extract the accelerometer power feature in the PPG frequency range.

    Parameters
    ----------
    f1: np.ndarray
        The frequency bins of the accelerometer signal.
    PSD_acc: np.ndarray
        The power spectral density of the accelerometer signal.
    f2: np.ndarray
        The frequency bins of the PPG signal.
    PSD_ppg: np.ndarray
        The power spectral density of the PPG signal.

    Returns
    -------
    np.ndarray
        The accelerometer power feature in the PPG frequency range
    """
    
    # Find the index of the maximum PSD value in the PPG signal
    max_PPG_psd_idx = np.argmax(PSD_ppg, axis=1)
    max_PPG_freq_psd = f2[max_PPG_psd_idx]
    
    # Find the neighboring indices of the maximum PSD value in the PPG signal
    df_idx = np.column_stack((max_PPG_psd_idx - 1, max_PPG_psd_idx, max_PPG_psd_idx + 1))    
    
    # Find the index of the closest frequency in the accelerometer signal to the first harmonic of the PPG frequency
    corr_acc_psd_fh_idx = np.argmin(np.abs(f1[:, None] - max_PPG_freq_psd*2), axis=0)
    fh_idx = np.column_stack((corr_acc_psd_fh_idx - 1, corr_acc_psd_fh_idx, corr_acc_psd_fh_idx + 1))   
    
    # Compute the power in the ranges corresponding to the PPG frequency
    acc_power_PPG_range = (
        np.trapz(PSD_acc[np.arange(PSD_acc.shape[0])[:, None], df_idx], f1[df_idx], axis=1) +
        np.trapz(PSD_acc[np.arange(PSD_acc.shape[0])[:, None], fh_idx], f1[fh_idx], axis=1)
    )

    # Compute the total power across the entire frequency range
    acc_power_total = np.trapz(PSD_acc, f1)
    
    # Compute the power ratio of the accelerometer signal in the PPG frequency range
    acc_power_ratio = acc_power_PPG_range / acc_power_total
    
    return acc_power_ratio

def extract_accelerometer_feature(
        acc_windowed: np.ndarray, 
        ppg_windowed: np.ndarray,
        config: PulseRateConfig
    ) -> pd.DataFrame:
    """
    Extract accelerometer features from the accelerometer signal in the PPG frequency range.
    
    Parameters
    ----------    
    acc_windowed: np.ndarray
        The dataframe containing the windowed accelerometer signal

    ppg_windowed: np.ndarray
        The dataframe containing the corresponding windowed ppg signal
    
    config: PulseRateConfig
        The configuration object containing the parameters for the feature extraction

    Returns
    -------
    pd.DataFrame
        The dataframe with the relative power accelerometer feature.
    """
    
    if config.sensor not in ['imu', 'ppg']:
        raise ValueError("Sensor not recognized.")
    
    d_freq = {}
    d_psd = {}
    for sensor in ['imu', 'ppg']:
        config.set_sensor(sensor)

        if sensor == 'imu':
            windows = acc_windowed
        else:
            windows = ppg_windowed

        window_type = hann(config.window_length_welch, sym = True)
        d_freq[sensor], d_psd[sensor] = welch(
            windows,
            fs=config.sampling_frequency,
            window=window_type,
            noverlap=config.overlap_welch_window,
            nfft=config.nfft,
            detrend=False,
            axis=1
        )

    d_psd['imu'] = np.sum(d_psd['imu'], axis=2)  # Sum the PSDs of the three axes

    acc_power_ratio = extract_acc_power_feature(d_freq['imu'], d_psd['imu'], d_freq['ppg'], d_psd['ppg'])

    return pd.DataFrame(acc_power_ratio, columns=['acc_power_ratio'])





import os
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from dateutil import parser
from typing import List, Tuple

import tsdf
from tsdf import TSDFMetadata

from paradigma.constants import DataColumns, TimeUnit


def parse_iso8601_to_datetime(date_str):
    return parser.parse(date_str)


def format_datetime_to_iso8601(datetime):
    return datetime.strftime("%Y-%m-%dT%H:%M:%S") + "Z"


def get_end_iso8601(start_iso8601, window_length_seconds):
    start_date = parser.parse(start_iso8601)
    end_date = start_date + timedelta(seconds=window_length_seconds)
    return format_datetime_to_iso8601(end_date)


def write_np_data(
    metadata_time: TSDFMetadata,
    np_array_time: np.ndarray, 
    metadata_values: TSDFMetadata,
    np_array_values: np.ndarray,
    output_path: str,
    output_filename: str,
):
    """
    Write the numpy arrays to binary files and store the metadata.

    Parameters
    ----------
    metadata_time : TSDFMetadata
        Metadata for the time column.
    np_array_time : np.ndarray
        The numpy array for the time column.
    metadata_values : TSDFMetadata
        Metadata for the samples columns.
    np_array_values : np.ndarray
        The numpy array for the samples columns.
    output_path : str
        The path where the files will be stored.
    output_filename : str
        The filename for the metadata.

    """
    
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    # TODO: improve the way the metadata is stored at a different location
    metadata_time.file_dir_path = output_path
    metadata_values.file_dir_path = output_path

    # store binaries and metadata
    time_tsdf = tsdf.write_binary_file(file_dir=output_path, file_name=metadata_time.file_name, data=np_array_time, metadata=metadata_time.get_plain_tsdf_dict_copy())

    samples_tsdf = tsdf.write_binary_file(file_dir=output_path, file_name=metadata_values.file_name, data=np_array_values, metadata=metadata_values.get_plain_tsdf_dict_copy())

    tsdf.write_metadata([time_tsdf, samples_tsdf], output_filename)


def write_df_data(
    metadata_time: TSDFMetadata,
    metadata_values: TSDFMetadata,
    output_path: str,
    output_filename: str,
    df: pd.DataFrame,
):
    """
    Write the Pandas DataFrame to binary files and store the metadata.

    Parameters
    ----------
    metadata_time : TSDFMetadata
        Metadata for the time column.
    metadata_values : TSDFMetadata
        Metadata for the samples columns.
    output_path : str
        The path where the files will be stored.
    output_filename : str
        The filename for the metadata.
    df : pd.DataFrame
        The DataFrame to be stored.

    """
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    # Make sure the iso8601 format is correctly set
    # TODO: this should be properly validated in the tsdf library instead
    start_date = parser.parse(metadata_time.start_iso8601)
    metadata_time.start_iso8601 = format_datetime_to_iso8601(start_date)
    end_date = parser.parse(metadata_time.end_iso8601)
    metadata_time.end_iso8601 = format_datetime_to_iso8601(end_date)
    start_date = parser.parse(metadata_values.start_iso8601)
    metadata_values.start_iso8601 = format_datetime_to_iso8601(start_date)
    end_date = parser.parse(metadata_values.end_iso8601)
    metadata_values.end_iso8601 = format_datetime_to_iso8601(end_date)

    # TODO: improve the way the metadata is stored at a different location
    metadata_time.file_dir_path = output_path
    metadata_values.file_dir_path = output_path

    # store binaries and metadata
    tsdf.write_dataframe_to_binaries(output_path, df, [metadata_time, metadata_values])
    tsdf.write_metadata([metadata_time, metadata_values], output_filename)


def read_metadata(
    input_path: str, meta_filename: str, time_filename: str, values_filename: str
) -> Tuple[TSDFMetadata, TSDFMetadata]:
    metadata_dict = tsdf.load_metadata_from_path(
        os.path.join(input_path, meta_filename)
    )
    metadata_time = metadata_dict[time_filename]
    metadata_values = metadata_dict[values_filename]
    return metadata_time, metadata_values


def load_tsdf_dataframe(path_to_data, prefix, meta_suffix='meta.json', time_suffix='time.bin', values_suffix='values.bin'):
    meta_filename = f"{prefix}_{meta_suffix}"
    time_filename = f"{prefix}_{time_suffix}"
    values_filename = f"{prefix}_{values_suffix}"

    metadata_time, metadata_values = read_metadata(path_to_data, meta_filename, time_filename, values_filename)
    df = tsdf.load_dataframe_from_binaries([metadata_time, metadata_values], tsdf.constants.ConcatenationType.columns)

    return df, metadata_time, metadata_values


def load_metadata_list(
    dir_path: str, meta_filename: str, filenames: List[str]
) -> List[TSDFMetadata]:
    """
    Load the metadata objects from a metadata file according to the specified binaries.

    Parameters
    ----------
    dir_path : str
        The dir path where the metadata file is stored.
    meta_filename : str
        The filename of the metadata file.
    filenames : List[str]
        The list of binary files of which the metadata files need to be loaded
    
    """	
    metadata_dict = tsdf.load_metadata_from_path(
        os.path.join(dir_path, meta_filename)
    )
    metadata_list = []
    for filename in filenames:
        metadata_list.append(metadata_dict[filename])

    return metadata_list


def transform_time_array(
    time_array: pd.Series,
    input_unit_type: str,
    output_unit_type: str,
    start_time: float = 0.0,
) -> np.ndarray:
    """
    Transforms the time array to relative time (when defined in delta time) and scales the values.

    Parameters
    ----------
    time_array : pd.Series
        The time array to transform.
    input_unit_type : str
        The time unit type of the input time array.
    output_unit_type : str
        The time unit type of the output time array. ParaDigMa expects `TimeUnit.RELATIVE_S`.
    start_time : float, optional
        The start time of the time array in UNIX seconds (default is 0.0)

    Returns
    -------
    np.ndarray
        The transformed time array in seconds, with the specified time unit type.

    Notes
    -----
    - The function handles different time units (`TimeUnit.RELATIVE_MS`, `TimeUnit.RELATIVE_S`, `TimeUnit.ABSOLUTE_MS`, `TimeUnit.ABSOLUTE_S`, `TimeUnit.DIFFERENCE_MS`, `TimeUnit.DIFFERENCE_S`).
    - The transformation allows for scaling of the time array, converting between time unit types (e.g., relative, absolute, or difference).
    - When converting to `TimeUnit.RELATIVE_MS`, the function calculates the relative time starting from the provided or default start time.
    """
    input_units = input_unit_type.split('_')[-1].lower()
    output_units = output_unit_type.split('_')[-1].lower()

    if input_units == output_units:
        scale_factor = 1
    elif input_units == 's' and output_units == 'ms':
        scale_factor = 1e3
    elif input_units == 'ms' and output_units == 's':
        scale_factor = 1 / 1e3
    else:
        raise ValueError(f"Unsupported time units conversion: {input_units} to {output_units}")
    
    # Transform to relative time (`TimeUnit.RELATIVE_MS`) 
    if input_unit_type == TimeUnit.DIFFERENCE_MS or input_unit_type == TimeUnit.DIFFERENCE_S:
    # Convert a series of differences into cumulative sum to reconstruct original time series.
        time_array = np.cumsum(np.double(time_array))
    elif input_unit_type == TimeUnit.ABSOLUTE_MS or input_unit_type == TimeUnit.ABSOLUTE_S:
        # Set the start time if not provided.
        if np.isclose(start_time, 0.0, rtol=1e-09, atol=1e-09):
            start_time = time_array[0]
        # Convert absolute time stamps into a time series relative to start_time.
        time_array = (time_array - start_time) 

    # Transform the time array from `TimeUnit.RELATIVE_MS` to the specified time unit type
    if output_unit_type == TimeUnit.ABSOLUTE_MS or output_unit_type == TimeUnit.ABSOLUTE_S:
        # Converts time array to absolute time by adding the start time to each element.
        time_array = time_array + start_time
    elif output_unit_type == TimeUnit.DIFFERENCE_MS or output_unit_type == TimeUnit.DIFFERENCE_S:
        # Creates a new array starting with 0, followed by the differences between consecutive elements.
        time_array = np.diff(np.insert(time_array, 0, start_time))
    elif output_unit_type == TimeUnit.RELATIVE_MS or output_unit_type == TimeUnit.RELATIVE_S:
        # The array is already in relative format, do nothing.
        pass

    return time_array * scale_factor


def convert_units_accelerometer(data: np.ndarray, units: str) -> np.ndarray:
    """
    Convert acceleration data to g.

    Parameters
    ----------
    data : np.ndarray
        The acceleration data.

    units : str
        The unit of the data (currently supports g and m/s^2).

    Returns
    -------
    np.ndarray
        The acceleration data in g.

    """
    if units == "m/s^2":
        return data / 9.81
    elif units == "g":
        return data
    else:
        raise ValueError(f"Unsupported unit: {units}")
    

def convert_units_gyroscope(data: np.ndarray, units: str) -> np.ndarray:
    """
    Convert gyroscope data to deg/s.
    
    Parameters
    ----------
    data : np.ndarray
        The gyroscope data.
        
    units : str
        The unit of the data (currently supports deg/s and rad/s).
        
    Returns
    -------
    np.ndarray
        The gyroscope data in deg/s.
        
    """
    if units == "deg/s":
        return data
    elif units == "rad/s":
        return np.degrees(data)
    else:
        raise ValueError(f"Unsupported unit: {units}")
    

def invert_watch_side(df: pd.DataFrame, side: str, sensor='both') -> np.ndarray:
    """
    Invert the data based on the watch side.

    Parameters
    ----------
    df : pd.DataFrame
        The data.
    side : str
        The watch side (left or right).
    sensor: str
        The sensor(s) to invert: 'accelerometer', 'gyroscope', or 'both'

    Returns
    -------
    pd.DataFrame
        The inverted data.

    """
    if side not in ["left", "right"]:
        raise ValueError(f"Unsupported side: {side}")
    if sensor not in ['accelerometer', 'gyroscope', 'both']:
        raise ValueError(f"Unsupported sensor: {sensor}")

    elif side == "right":
        if sensor in ['gyroscope', 'both']:
            df[DataColumns.GYROSCOPE_Y] *= -1
            df[DataColumns.GYROSCOPE_Z] *= -1
        if sensor in ['accelerometer', 'both']:
            df[DataColumns.ACCELEROMETER_X] *= -1

    return df

def aggregate_parameter(parameter: np.ndarray, aggregate: str) -> np.ndarray:
    """
    Aggregate a parameter based on the specified method.
    
    Parameters
    ----------
    parameter : np.ndarray
        The parameter to aggregate.
        
    aggregate : str
        The aggregation method to apply.
        
    Returns
    -------
    np.ndarray
        The aggregated parameter.
    """
    if aggregate == 'mean':
        return np.mean(parameter)
    elif aggregate == 'median':
        return np.median(parameter)
    elif aggregate == 'mode':
        unique_values, counts = np.unique(parameter, return_counts=True)
        return unique_values[np.argmax(counts)]
    elif aggregate == '90p':
        return np.percentile(parameter, 90)
    elif aggregate == '95p':
        return np.percentile(parameter, 95)
    elif aggregate == '99p':
        return np.percentile(parameter, 99)
    elif aggregate == 'std':
        return np.std(parameter)
    else:
        raise ValueError(f"Invalid aggregation method: {aggregate}")

def merge_predictions_with_timestamps(
        df_ts: pd.DataFrame, 
        df_predictions: pd.DataFrame, 
        pred_proba_colname: str,
        window_length_s: float, 
        fs: int
    ) -> pd.DataFrame:
    """
    Merges prediction probabilities with timestamps by expanding overlapping windows
    into individual timestamps and averaging probabilities per unique timestamp.

    Parameters:
    ----------
    df_ts : pd.DataFrame
        DataFrame containing timestamps to be merged with predictions.
        Must include the timestamp column specified in `DataColumns.TIME`.

    df_predictions : pd.DataFrame
        DataFrame containing prediction windows with start times and probabilities.
        Must include:
        - A column for window start times (defined by `DataColumns.TIME`).
        - A column for prediction probabilities (defined by `DataColumns.PRED_GAIT_PROBA`).

    pred_proba_colname : str
        The column name for the prediction probabilities in `df_predictions`.

    window_length_s : float
        The length of the prediction window in seconds.

    fs : int
        The sampling frequency of the data.
        
    Returns:
    -------
    pd.DataFrame
        Updated `df_ts` with an additional column for averaged prediction probabilities.

    Steps:
    ------
    1. Expand prediction windows into individual timestamps using NumPy broadcasting.
    2. Flatten the timestamps and prediction probabilities into single arrays.
    3. Aggregate probabilities by unique timestamps using pandas `groupby`.
    4. Merge the aggregated probabilities with the input `df_ts`.

    Notes:
    ------
    - Rounding is applied to timestamps to mitigate floating-point inaccuracies.
    - Fully vectorized for speed and scalability, avoiding any row-wise operations.
    """
    # Step 1: Generate all timestamps for prediction windows using NumPy broadcasting
    window_length = int(window_length_s * fs)
    timestamps = (
        df_predictions[DataColumns.TIME].values[:, None] +
        np.arange(0, window_length) / fs
    )
    
    # Flatten timestamps and probabilities into a single array for efficient processing
    flat_timestamps = timestamps.ravel()
    flat_proba = np.repeat(
        df_predictions[pred_proba_colname].values,
        window_length
    )

    # Step 2: Create a DataFrame for expanded data
    expanded_df = pd.DataFrame({
        DataColumns.TIME: flat_timestamps,
        pred_proba_colname: flat_proba
    })

    # Step 3: Round timestamps and aggregate probabilities
    expanded_df[DataColumns.TIME] = expanded_df[DataColumns.TIME].round(2)
    mean_proba = expanded_df.groupby(DataColumns.TIME, as_index=False).mean()

    # Step 4: Round timestamps in `df_ts` and merge
    df_ts[DataColumns.TIME] = df_ts[DataColumns.TIME].round(2)
    df_ts = pd.merge(df_ts, mean_proba, how='left', on=DataColumns.TIME)
    df_ts = df_ts.dropna(subset=[pred_proba_colname])

    return df_ts


def select_hours(df: pd.DataFrame, select_hours_start: str, select_hours_end: str) -> pd.DataFrame:
    
    """
    Select hours of interest from the data to include in the aggregation step.

    Parameters
    ----------
    df : pd.DataFrame
        Input data.

    select_hours_start: str
        The start time of the selected hours in "HH:MM" format.
    
    select_hours_end: str
            The end time of the selected hours in "HH:MM" format.

    Returns
    -------
    pd.DataFrame
        The selected data.

    """

    select_hours_start = datetime.strptime(select_hours_start, '%H:%M').time() # convert to time object
    select_hours_end = datetime.strptime(select_hours_end, '%H:%M').time() 
    df_subset = df[df['time_dt'].dt.time.between(select_hours_start, select_hours_end)] # select the hours of interest

    return df_subset

def select_days(df: pd.DataFrame, min_hours_per_day: int) -> pd.DataFrame:

    """
    Select days of interest from the data to include in the aggregation step.

    Parameters
    ----------
    df : pd.DataFrame
        Input data with column 'time_dt' in which the date is stored.

    min_hours_per_day: int
        The minimum number of hours per day required for including the day in the aggregation step.


    Returns
    -------
    pd.DataFrame
        The selected data.

    """

    min_s_per_day = min_hours_per_day * 3600
    window_length_s = df['time_dt'].diff().dt.total_seconds()[1] # determine the length of the first window in seconds
    min_windows_per_day = min_s_per_day / window_length_s
    df_subset = df.groupby(df['time_dt'].dt.date).filter(lambda x: len(x) >= min_windows_per_day)

    return df_subset


import pandas as pd
import numpy as np
from pathlib import Path
from scipy import signal
from scipy.stats import gaussian_kde

from paradigma.classification import ClassifierPackage
from paradigma.constants import DataColumns
from paradigma.config import TremorConfig
from paradigma.feature_extraction import compute_mfccs, compute_power_in_bandwidth, compute_total_power, extract_frequency_peak, \
    extract_tremor_power
from paradigma.segmenting import tabulate_windows, WindowedDataExtractor
from paradigma.util import aggregate_parameter


def extract_tremor_features(df: pd.DataFrame, config: TremorConfig) -> pd.DataFrame:
    """
    This function groups sequences of timestamps into windows and subsequently extracts 
    tremor features from windowed gyroscope data.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing sensor data, which includes time and gyroscope data. The data should be
        structured with the necessary columns as specified in the `config`.

    config : TremorConfig
        Configuration object containing parameters for feature extraction, including column names for time, gyroscope data,
        as well as settings for windowing, and feature computation.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing extracted tremor features and a column corresponding to time.
    
    Notes
    -----
    - This function groups the data into windows based on timestamps.
    - The input DataFrame must include columns as specified in the `config` object for proper feature extraction.

    Raises
    ------
    ValueError
        If the input DataFrame does not contain the required columns as specified in the configuration or if any step in the feature extraction fails.
    """
    # group sequences of timestamps into windows
    windowed_cols = [DataColumns.TIME] + config.gyroscope_cols
    windowed_data = tabulate_windows(df, windowed_cols, config.window_length_s, config.window_step_length_s, config.sampling_frequency)

    extractor = WindowedDataExtractor(windowed_cols)

    # Extract the start time and gyroscope data from the windowed data
    idx_time = extractor.get_index(DataColumns.TIME)
    idx_gyro = extractor.get_slice(config.gyroscope_cols)

    # Extract data
    start_time = np.min(windowed_data[:, :, idx_time], axis=1)
    windowed_gyro = windowed_data[:, :, idx_gyro]

    df_features = pd.DataFrame(start_time, columns=[DataColumns.TIME])
    
    # transform the signals from the temporal domain to the spectral domain and extract tremor features
    df_spectral_features = extract_spectral_domain_features(windowed_gyro, config)

    # Combine spectral features with the start time
    df_features = pd.concat([df_features, df_spectral_features], axis=1)

    return df_features


def detect_tremor(df: pd.DataFrame, config: TremorConfig, full_path_to_classifier_package: str | Path) -> pd.DataFrame:
    """
    Detects tremor in the input DataFrame using a pre-trained classifier and applies a threshold to the predicted probabilities.

    This function performs the following steps:
    1. Loads the pre-trained classifier and scaling parameters from the provided directory.
    2. Scales the relevant features in the input DataFrame (`df`) using the loaded scaling parameters.
    3. Makes predictions using the classifier to estimate the probability of tremor.
    4. Applies a threshold to the predicted probabilities to classify whether tremor is detected or not.
    5. Checks for rest tremor by verifying the frequency of the peak and below tremor power.
    6. Adds the predicted probabilities and the classification result to the DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing extracted tremor features. The DataFrame must include
        the necessary columns as specified in the classifier's feature names.

    config : TremorConfig
        Configuration object containing settings for tremor detection, including the frequency range for rest tremor.

    full_path_to_classifier_package : str | Path
        The path to the directory containing the classifier file, threshold value, scaler parameters, and other necessary input
        files for tremor detection.

    Returns
    -------
    pd.DataFrame
        The input DataFrame (`df`) with two additional columns:
        - `PRED_TREMOR_PROBA`: Predicted probability of tremor based on the classifier.
        - `PRED_TREMOR_LOGREG`: Binary classification result (True for tremor, False for no tremor), based on the threshold applied to `PRED_TREMOR_PROBA`.
        - `PRED_TREMOR_CHECKED`: Binary classification result (True for tremor, False for no tremor), after performing extra checks for rest tremor on `PRED_TREMOR_LOGREG`.
        - `PRED_ARM_AT_REST`: Binary classification result (True for arm at rest or stable posture, False for significant arm movement), based on the power below tremor.

    Notes
    -----
    - The threshold used to classify tremor is loaded from a file and applied to the predicted probabilities.

    Raises
    ------
    FileNotFoundError
        If the classifier, scaler, or threshold files are not found at the specified paths.
    ValueError
        If the DataFrame does not contain the expected features for prediction or if the prediction fails.

    """

    # Load the classifier package
    clf_package = ClassifierPackage.load(full_path_to_classifier_package)

    # Set classifier
    clf = clf_package.classifier
    feature_names_scaling = clf_package.scaler.feature_names_in_
    feature_names_predictions = clf.feature_names_in_

    # Apply scaling to relevant columns
    scaled_features = clf_package.transform_features(df.loc[:, feature_names_scaling])

    # Replace scaled features in a copy of the relevant features for prediction
    X = df.loc[:, feature_names_predictions].copy()
    X.loc[:, feature_names_scaling] = scaled_features

    # Get the tremor probability 
    df[DataColumns.PRED_TREMOR_PROBA] = clf_package.predict_proba(X)

    # Make prediction based on pre-defined threshold
    df[DataColumns.PRED_TREMOR_LOGREG] = (df[DataColumns.PRED_TREMOR_PROBA] >= clf_package.threshold).astype(int)

    # Perform extra checks for rest tremor 
    peak_check = (df['freq_peak'] >= config.fmin_rest_tremor) & (df['freq_peak']<=config.fmax_rest_tremor) # peak within 3-7 Hz
    df[DataColumns.PRED_ARM_AT_REST] = (df['below_tremor_power'] <= config.movement_threshold).astype(int) # arm at rest or in stable posture
    df[DataColumns.PRED_TREMOR_CHECKED] = ((df[DataColumns.PRED_TREMOR_LOGREG]==1) & (peak_check==True) & (df[DataColumns.PRED_ARM_AT_REST] == True)).astype(int)
    
    return df

def aggregate_tremor(df: pd.DataFrame, config: TremorConfig):
    """
    Quantifies the amount of tremor time and tremor power, aggregated over all windows in the input dataframe.
    Tremor time is calculated as the number of the detected tremor windows, as percentage of the number of windows 
    without significant non-tremor movement (at rest). For tremor power the following aggregates are derived:
    the median, mode and percentile of tremor power specified in the configuration object. 
    
    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing the tremor predictions and computed tremor power.
        The DataFrame must also contain a datatime column ('time_dt').

    config : TremorConfig
        Configuration object containing the percentile for aggregating tremor power.

    Returns
    -------
    dict
        A dictionary with the aggregated tremor time and tremor power measures, as well as the number of valid days,
        the total number of windows, and the number of windows at rest available in the input dataframe.

    Notes
    -----
    - Tremor power is converted to log scale, after adding a constant of 1, so that zero tremor power
    corresponds to a value of 0 in log scale.
    - The modal tremor power is computed based on gaussian kernel density estimation.
  
    """
    nr_valid_days = df['time_dt'].dt.date.unique().size # number of valid days in the input dataframe
    nr_windows_total = df.shape[0] # number of windows in the input dataframe

    # remove windows with detected non-tremor arm movements to control for the amount of arm activities performed
    df_filtered = df.loc[df.pred_arm_at_rest == 1]
    nr_windows_rest = df_filtered.shape[0] # number of windows without non-tremor arm movement

    if nr_windows_rest == 0: # if no windows without non-tremor arm movement are detected
        raise Warning('No windows without non-tremor arm movement are detected.')

    # calculate tremor time
    n_windows_tremor = np.sum(df_filtered['pred_tremor_checked'])
    perc_windows_tremor = n_windows_tremor / nr_windows_rest * 100 # as percentage of total measured time without non-tremor arm movement

    aggregated_tremor_power = {} # initialize dictionary to store aggregated tremor power measures
    
    if n_windows_tremor == 0: # if no tremor is detected, the tremor power measures are set to NaN

        aggregated_tremor_power['median_tremor_power'] = np.nan
        aggregated_tremor_power['modal_tremor_power'] = np.nan
        aggregated_tremor_power['90p_tremor_power'] = np.nan

    else:
        
        # calculate aggregated tremor power measures
        tremor_power = df_filtered.loc[df_filtered['pred_tremor_checked'] == 1, 'tremor_power']
        tremor_power = np.log10(tremor_power+1) # convert to log scale
        
        for aggregate in config.aggregates_tremor_power:
            aggregate_name = f"{aggregate}_tremor_power"
            if aggregate == 'mode':
                # calculate modal tremor power
                bin_edges = np.linspace(0, 6, 301)
                kde = gaussian_kde(tremor_power)
                kde_values = kde(bin_edges)
                max_index = np.argmax(kde_values)
                aggregated_tremor_power['modal_tremor_power'] = bin_edges[max_index]
            else: # calculate te other aggregates (e.g. median and 90th percentile) of tremor power
                aggregated_tremor_power[aggregate_name] = aggregate_parameter(tremor_power, aggregate)
    
    # store aggregates in json format
    d_aggregates = {
        'metadata': {
            'nr_valid_days': nr_valid_days,
            'nr_windows_total': nr_windows_total,
            'nr_windows_rest': nr_windows_rest
        },
        'aggregated_tremor_measures': {
            'perc_windows_tremor': perc_windows_tremor,
            'median_tremor_power': aggregated_tremor_power['median_tremor_power'],
            'modal_tremor_power': aggregated_tremor_power['modal_tremor_power'],
            '90p_tremor_power': aggregated_tremor_power['90p_tremor_power']
        }
    }

    return d_aggregates


def extract_spectral_domain_features(data: np.ndarray, config) -> pd.DataFrame:
    """
    Compute spectral domain features from the gyroscope data.

    This function computes Mel-frequency cepstral coefficients (MFCCs), the frequency of the peak, 
    the tremor power, and the below tremor power based on the total power spectral density of the windowed gyroscope data.

    Parameters
    ----------
    data : numpy.ndarray
        A 2D numpy array where each row corresponds to a window of gyroscope data.
    config : object
        Configuration object containing settings such as sampling frequency, window type, 
        and MFCC parameters.
    
    Returns
    -------
    pd.DataFrame
        The feature dataframe containing the extracted spectral features, including 
        MFCCs, the frequency of the peak, the tremor power and below tremor power for each window.
        
    """

    # Initialize a dictionary to hold the results
    feature_dict = {}

    # Initialize parameters
    sampling_frequency = config.sampling_frequency
    segment_length_psd_s = config.segment_length_psd_s
    segment_length_spectrogram_s = config.segment_length_spectrogram_s
    overlap_fraction = config.overlap_fraction
    spectral_resolution = config.spectral_resolution
    window_type = 'hann'

    # Compute the power spectral density
    segment_length_n = sampling_frequency * segment_length_psd_s
    overlap_n = segment_length_n * overlap_fraction
    window = signal.get_window(window_type, segment_length_n, fftbins=False)
    nfft = sampling_frequency / spectral_resolution

    freqs, psd = signal.welch(
        x=data, 
        fs=sampling_frequency, 
        window=window, 
        nperseg=segment_length_n,
        noverlap=overlap_n, 
        nfft=nfft, 
        detrend=False, 
        scaling='density',
        axis=1
    )

    # Compute the spectrogram
    segment_length_n = sampling_frequency * segment_length_spectrogram_s
    overlap_n = segment_length_n * overlap_fraction
    window = signal.get_window(window_type, segment_length_n)

    f, t, S1 = signal.stft(
        x=data, 
        fs=sampling_frequency, 
        window=window, 
        nperseg=segment_length_n, 
        noverlap=overlap_n,
        boundary=None,
        axis=1
    )

    # Compute total power in the PSD and the total spectrogram (summed over the three axes)
    total_psd = compute_total_power(psd)
    total_spectrogram = np.sum(np.abs(S1)*sampling_frequency, axis=2)

    # Compute the MFCC's
    config.mfcc_low_frequency = config.fmin_mfcc
    config.mfcc_high_frequency = config.fmax_mfcc
    config.mfcc_n_dct_filters = config.n_dct_filters_mfcc
    config.mfcc_n_coefficients = config.n_coefficients_mfcc

    mfccs = compute_mfccs(
        total_power_array=total_spectrogram,
        config=config,
        total_power_type='spectrogram',
        rounding_method='round',
        multiplication_factor=1
    )

    # Combine the MFCCs into the features DataFrame
    mfcc_colnames = [f'mfcc_{x}' for x in range(1, config.mfcc_n_coefficients + 1)]
    for i, colname in enumerate(mfcc_colnames):
        feature_dict[colname] = mfccs[:, i]

    # Compute the frequency of the peak, non-tremor power and tremor power
    feature_dict['freq_peak'] = extract_frequency_peak(freqs, total_psd, config.fmin_peak_search, config.fmax_peak_search)
    feature_dict['below_tremor_power'] = compute_power_in_bandwidth(freqs, total_psd, config.fmin_below_rest_tremor, config.fmax_below_rest_tremor, 
                                                                include_max=False, spectral_resolution=config.spectral_resolution, 
                                                                cumulative_sum_method='sum')
    feature_dict['tremor_power'] = extract_tremor_power(freqs, total_psd, config.fmin_rest_tremor, config.fmax_rest_tremor)

    return pd.DataFrame(feature_dict)


import numpy as np
import pandas as pd
from scipy.signal import periodogram
from typing import List, Tuple

from paradigma.classification import ClassifierPackage
from paradigma.constants import DataColumns
from paradigma.config import GaitConfig
from paradigma.feature_extraction import pca_transform_gyroscope, compute_angle, remove_moving_average_angle, \
    extract_angle_extremes, compute_range_of_motion, compute_peak_angular_velocity, compute_statistics, \
    compute_std_euclidean_norm, compute_power_in_bandwidth, compute_dominant_frequency, compute_mfccs, \
    compute_total_power
from paradigma.segmenting import tabulate_windows, create_segments, discard_segments, categorize_segments, WindowedDataExtractor
from paradigma.util import aggregate_parameter


def extract_gait_features(
        df: pd.DataFrame,
        config: GaitConfig
    ) -> pd.DataFrame:
    """
    Extracts gait features from accelerometer and gravity sensor data in the input DataFrame by computing temporal and spectral features.

    This function performs the following steps:
    1. Groups sequences of timestamps into windows, using accelerometer and gravity data.
    2. Computes temporal domain features such as mean and standard deviation for accelerometer and gravity data.
    3. Transforms the signals from the temporal domain to the spectral domain using the Fast Fourier Transform (FFT).
    4. Computes spectral domain features for the accelerometer data.
    5. Combines both temporal and spectral features into a final DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing gait data, which includes time, accelerometer, and gravity sensor data. The data should be
        structured with the necessary columns as specified in the `config`.

    onfig : GaitConfig
        Configuration object containing parameters for feature extraction, including column names for time, accelerometer data, and
        gravity data, as well as settings for windowing, and feature computation.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing extracted gait features, including temporal and spectral domain features. The DataFrame will have
        columns corresponding to time, statistical features of the accelerometer and gravity data, and spectral features of the
        accelerometer data.
    
    Notes
    -----
    - This function groups the data into windows based on timestamps and applies Fast Fourier Transform to compute spectral features.
    - The temporal features are extracted from the accelerometer and gravity data, and include statistics like mean and standard deviation.
    - The input DataFrame must include columns as specified in the `config` object for proper feature extraction.

    Raises
    ------
    ValueError
        If the input DataFrame does not contain the required columns as specified in the configuration or if any step in the feature extraction fails.
    """
    # Group sequences of timestamps into windows
    windowed_cols = [DataColumns.TIME] + config.accelerometer_cols + config.gravity_cols
    windowed_data = tabulate_windows(
        df=df, 
        columns=windowed_cols,
        window_length_s=config.window_length_s,
        window_step_length_s=config.window_step_length_s,
        fs=config.sampling_frequency
    )

    extractor = WindowedDataExtractor(windowed_cols)

    idx_time = extractor.get_index(DataColumns.TIME)
    idx_acc = extractor.get_slice(config.accelerometer_cols)
    idx_grav = extractor.get_slice(config.gravity_cols)

    # Extract data
    start_time = np.min(windowed_data[:, :, idx_time], axis=1)
    windowed_acc = windowed_data[:, :, idx_acc]
    windowed_grav = windowed_data[:, :, idx_grav]

    df_features = pd.DataFrame(start_time, columns=[DataColumns.TIME])
    
    # Compute statistics of the temporal domain signals (mean, std) for accelerometer and gravity
    df_temporal_features = extract_temporal_domain_features(
        config=config, 
        windowed_acc=windowed_acc,
        windowed_grav=windowed_grav,
        grav_stats=['mean', 'std']
    )

    # Combine temporal features with the start time
    df_features = pd.concat([df_features, df_temporal_features], axis=1)

    # Transform the accelerometer data to the spectral domain using FFT and extract spectral features
    df_spectral_features = extract_spectral_domain_features(
        config=config, 
        sensor='accelerometer', 
        windowed_data=windowed_acc
    )

    # Combine the spectral features with the previously computed temporal features
    df_features = pd.concat([df_features, df_spectral_features], axis=1)

    return df_features


def detect_gait(
        df: pd.DataFrame, 
        clf_package: ClassifierPackage, 
        parallel: bool=False
    ) -> pd.Series:
    """
    Detects gait activity in the input DataFrame using a pre-trained classifier and applies a threshold to classify results.

    This function performs the following steps:
    1. Loads the pre-trained classifier and scaling parameters from the specified directory.
    2. Scales the relevant features in the input DataFrame (`df`) using the loaded scaling parameters.
    3. Predicts the probability of gait activity for each sample in the DataFrame using the classifier.
    4. Applies a threshold to the predicted probabilities to determine whether gait activity is present.
    5. Returns predicted probabilities

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing features extracted from gait data. It must include the necessary columns 
        as specified in the classifier's feature names.

    clf_package : ClassifierPackage
        The pre-trained classifier package containing the classifier, threshold, and scaler.

    parallel : bool, optional, default=False
        If `True`, enables parallel processing during classification. If `False`, the classifier uses a single core.

    Returns
    -------
    pd.Series
        A Series containing the predicted probabilities of gait activity for each sample in the input DataFrame.
    """
    # Set classifier
    clf = clf_package.classifier
    if not parallel and hasattr(clf, 'n_jobs'):
        clf.n_jobs = 1

    feature_names_scaling = clf_package.scaler.feature_names_in_
    feature_names_predictions = clf.feature_names_in_

    # Apply scaling to relevant columns
    scaled_features = clf_package.transform_features(df.loc[:, feature_names_scaling])

    # Replace scaled features in a copy of the relevant features for prediction
    X = df.loc[:, feature_names_predictions].copy()
    X.loc[:, feature_names_scaling] = scaled_features

    # Make prediction and add the probability of gait activity to the DataFrame
    pred_gait_proba_series = clf_package.predict_proba(X)

    return pred_gait_proba_series


def extract_arm_activity_features(
        df: pd.DataFrame, 
        config: GaitConfig,
    ) -> pd.DataFrame:
    """
    Extract features related to arm activity from a time-series DataFrame.

    This function processes a DataFrame containing accelerometer, gravity, and gyroscope signals, 
    and extracts features related to arm activity by performing the following steps:
    1. Computes the angle and velocity from gyroscope data.
    2. Filters the data to include only predicted gait segments.
    3. Groups the data into segments based on consecutive timestamps and pre-specified gaps.
    4. Removes segments that do not meet predefined criteria.
    5. Creates fixed-length windows from the time series data.
    6. Extracts angle-related features, temporal domain features, and spectral domain features.

    Parameters
    ----------
    df: pd.DataFrame
        The input DataFrame containing accelerometer, gravity, and gyroscope data of predicted gait.

    config : ArmActivityFeatureExtractionConfig
        Configuration object containing column names and parameters for feature extraction.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the extracted arm activity features, including angle, velocity, 
        temporal, and spectral features.
    """
    # Group consecutive timestamps into segments, with new segments starting after a pre-specified gap
    df[DataColumns.SEGMENT_NR] = create_segments(
        time_array=df[DataColumns.TIME], 
        max_segment_gap_s=config.max_segment_gap_s
    )

    # Remove segments that do not meet predetermined criteria
    df = discard_segments(
        df=df,
        segment_nr_colname=DataColumns.SEGMENT_NR,
        min_segment_length_s=config.min_segment_length_s,
        fs=config.sampling_frequency,
        format='timestamps'
    )

    # Create windows of fixed length and step size from the time series per segment
    windowed_data = []
    df_grouped = df.groupby(DataColumns.SEGMENT_NR)
    windowed_cols = (
        [DataColumns.TIME] + 
        config.accelerometer_cols + 
        config.gravity_cols + 
        config.gyroscope_cols
    )

    # Collect windows from all segments in a list for faster concatenation
    for _, group in df_grouped:
        windows = tabulate_windows(
            df=group, 
            columns=windowed_cols,
            window_length_s=config.window_length_s,
            window_step_length_s=config.window_step_length_s,
            fs=config.sampling_frequency
        )
        if len(windows) > 0:  # Skip if no windows are created
            windowed_data.append(windows)

    # If no windows were created, raise an error
    if not windowed_data:
        print("No windows were created from the given data.")
        return pd.DataFrame()

    # Concatenate the windows into one array at the end
    windowed_data = np.concatenate(windowed_data, axis=0)

    # Slice columns for accelerometer, gravity, gyroscope, angle, and velocity
    extractor = WindowedDataExtractor(windowed_cols)

    idx_time = extractor.get_index(DataColumns.TIME)
    idx_acc = extractor.get_slice(config.accelerometer_cols)
    idx_grav = extractor.get_slice(config.gravity_cols)
    idx_gyro = extractor.get_slice(config.gyroscope_cols)

    # Extract data
    start_time = np.min(windowed_data[:, :, idx_time], axis=1)
    windowed_acc = windowed_data[:, :, idx_acc]
    windowed_grav = windowed_data[:, :, idx_grav]
    windowed_gyro = windowed_data[:, :, idx_gyro]

    # Initialize DataFrame for features
    df_features = pd.DataFrame(start_time, columns=[DataColumns.TIME])

    # Extract temporal domain features (e.g., mean, std for accelerometer and gravity)
    df_temporal_features = extract_temporal_domain_features(
        config=config, 
        windowed_acc=windowed_acc, 
        windowed_grav=windowed_grav, 
        grav_stats=['mean', 'std']
    )
    df_features = pd.concat([df_features, df_temporal_features], axis=1)

    # Extract spectral domain features for accelerometer and gyroscope signals
    for sensor_name, windowed_sensor in zip(['accelerometer', 'gyroscope'], [windowed_acc, windowed_gyro]):
        df_spectral_features = extract_spectral_domain_features(
            config=config, 
            sensor=sensor_name, 
            windowed_data=windowed_sensor
        )
        df_features = pd.concat([df_features, df_spectral_features], axis=1)

    return df_features


def filter_gait(
        df: pd.DataFrame, 
        clf_package: ClassifierPackage, 
        parallel: bool=False
    ) -> pd.Series:
    """
    Filters gait data to identify windows with no other arm activity using a pre-trained classifier.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing features extracted from gait data.
    clf_package: ClassifierPackage
        The pre-trained classifier package containing the classifier, threshold, and scaler.
    parallel : bool, optional, default=False
        If `True`, enables parallel processing.

    Returns
    -------
    pd.Series
        A Series containing the predicted probabilities.
    """
    if df.shape[0] == 0:
        raise ValueError("No data found in the input DataFrame.")
    
    # Set classifier
    clf = clf_package.classifier
    if not parallel and hasattr(clf, 'n_jobs'):
        clf.n_jobs = 1

    feature_names_scaling = clf_package.scaler.feature_names_in_
    feature_names_predictions = clf.feature_names_in_

    # Apply scaling to relevant columns
    scaled_features = clf_package.transform_features(df.loc[:, feature_names_scaling])

    # Replace scaled features in a copy of the relevant features for prediction
    X = df.loc[:, feature_names_predictions].copy()
    X.loc[:, feature_names_scaling] = scaled_features

    # Make predictions
    pred_no_other_arm_activity_proba_series = clf_package.predict_proba(X)

    return pred_no_other_arm_activity_proba_series


def quantify_arm_swing(
        df: pd.DataFrame, 
        fs: int,
        filtered: bool = False,
        max_segment_gap_s: float = 1.5,
        min_segment_length_s: float = 1.5
    ) -> Tuple[dict[str, pd.DataFrame], dict]:
    """
    Quantify arm swing parameters for segments of motion based on gyroscope data.

    Parameters
    ----------
    df : pd.DataFrame
        A DataFrame containing the raw sensor data of predicted gait timestamps. Should include a column
        for predicted no other arm activity based on a fitted threshold if filtered is True.

    fs : int
        The sampling frequency of the sensor data.

    filtered : bool, optional, default=True
        If `True`, the gyroscope data is filtered to only include predicted no other arm activity.

    max_segment_gap_s : float, optional, default=1.5
        The maximum gap in seconds between consecutive timestamps to group them into segments.
    
    min_segment_length_s : float, optional, default=1.5
        The minimum length in seconds for a segment to be considered valid.

    Returns
    -------
    Tuple[pd.DataFrame, dict]
        A tuple containing a dataframe with quantified arm swing parameters and a dictionary containing 
        metadata for each segment.
    """
    # Group consecutive timestamps into segments, with new segments starting after a pre-specified gap.
    # Segments are made based on predicted gait
    df[DataColumns.SEGMENT_NR] = create_segments(
        time_array=df[DataColumns.TIME], 
        max_segment_gap_s=max_segment_gap_s
    )

    # Segment category is determined based on predicted gait, hence it is set
    # before filtering the DataFrame to only include predicted no other arm activity
    df[DataColumns.SEGMENT_CAT] = categorize_segments(df=df, fs=fs)

    # Remove segments that do not meet predetermined criteria
    df = discard_segments(
        df=df,
        segment_nr_colname=DataColumns.SEGMENT_NR,
        min_segment_length_s=min_segment_length_s,
        fs=fs,
        format='timestamps'
    )

    if df.empty:
        raise ValueError("No segments found in the input data after discarding segments of invalid shape.")

    # If no arm swing data is remaining, return an empty dictionary
    if filtered and df.loc[df[DataColumns.PRED_NO_OTHER_ARM_ACTIVITY]==1].empty:
        raise ValueError("No gait without other arm activities to quantify.")
    elif filtered:
        # Filter the DataFrame to only include predicted no other arm activity (1)
        df = df.loc[df[DataColumns.PRED_NO_OTHER_ARM_ACTIVITY]==1].reset_index(drop=True)

        # Group consecutive timestamps into segments of filtered gait
        df[DataColumns.SEGMENT_NR] = create_segments(
            time_array=df[DataColumns.TIME], 
            max_segment_gap_s=max_segment_gap_s
        )

        # Remove segments that do not meet predetermined criteria
        df = discard_segments(
            df=df,
            segment_nr_colname=DataColumns.SEGMENT_NR,
            min_segment_length_s=min_segment_length_s,
            fs=fs,
        )

        if df.empty:
            raise ValueError("No filtered gait segments found in the input data after discarding segments of invalid shape.")

    arm_swing_quantified = []
    segment_meta = {
        'aggregated': {
            'all': {
                'duration_s': len(df[DataColumns.TIME]) / fs
            },
        },
        'per_segment': {}
    }

    # PCA is fitted on only predicted gait without other arm activity if filtered, otherwise
    # it is fitted on the entire gyroscope data
    df[DataColumns.VELOCITY] = pca_transform_gyroscope(
        df=df,
        y_gyro_colname=DataColumns.GYROSCOPE_Y,
        z_gyro_colname=DataColumns.GYROSCOPE_Z,
    )

    # Group and process segments
    for segment_nr, group in df.groupby(DataColumns.SEGMENT_NR, sort=False):
        segment_cat = group[DataColumns.SEGMENT_CAT].iloc[0]
        time_array = group[DataColumns.TIME].to_numpy()
        velocity_array = group[DataColumns.VELOCITY].to_numpy()

        # Integrate the angular velocity to obtain an estimation of the angle
        angle_array = compute_angle(
            time_array=time_array,
            velocity_array=velocity_array,
        )

        # Detrend angle using moving average
        angle_array = remove_moving_average_angle(
            angle_array=angle_array,
            fs=fs,
        )

        segment_meta['per_segment'][segment_nr] = {
            'start_time_s': time_array.min(),
            'end_time_s': time_array.max(),
            'duration_s': len(angle_array) / fs,
            DataColumns.SEGMENT_CAT: segment_cat
        }

        if angle_array.size > 0:  
            angle_extrema_indices, _, _ = extract_angle_extremes(
                angle_array=angle_array,
                sampling_frequency=fs,
                max_frequency_activity=1.75
            )

            if len(angle_extrema_indices) > 1:  # Requires at minimum 2 peaks
                try:
                    rom = compute_range_of_motion(
                        angle_array=angle_array,
                        extrema_indices=angle_extrema_indices,
                    )
                except Exception as e:
                    # Handle the error, set RoM to NaN, and log the error
                    print(f"Error computing range of motion for segment {segment_nr}: {e}")
                    rom = np.array([np.nan])

                try:
                    pav = compute_peak_angular_velocity(
                        velocity_array=velocity_array,
                        angle_extrema_indices=angle_extrema_indices
                    )
                except Exception as e:
                    # Handle the error, set pav to NaN, and log the error
                    print(f"Error computing peak angular velocity for segment {segment_nr}: {e}")
                    pav = np.array([np.nan])

                df_params_segment = pd.DataFrame({
                    DataColumns.SEGMENT_NR: segment_nr,
                    DataColumns.SEGMENT_CAT: segment_cat,
                    DataColumns.RANGE_OF_MOTION: rom,
                    DataColumns.PEAK_VELOCITY: pav
                })

                arm_swing_quantified.append(df_params_segment)

    # Combine segment categories
    segment_categories = set([segment_meta['per_segment'][x][DataColumns.SEGMENT_CAT] for x in segment_meta['per_segment'].keys()])
    for segment_cat in segment_categories:
        segment_meta['aggregated'][segment_cat] = {
            'duration_s': sum([segment_meta['per_segment'][x]['duration_s'] for x in segment_meta['per_segment'].keys() if segment_meta['per_segment'][x][DataColumns.SEGMENT_CAT] == segment_cat])
        }

    arm_swing_quantified = pd.concat(arm_swing_quantified, ignore_index=True)
            
    return arm_swing_quantified, segment_meta


def aggregate_arm_swing_params(df_arm_swing_params: pd.DataFrame, segment_meta: dict, aggregates: List[str] = ['median']) -> dict:
    """
    Aggregate the quantification results for arm swing parameters.
    
    Parameters
    ----------
    df_arm_swing_params : pd.DataFrame
        A dataframe containing the arm swing parameters to be aggregated

    segment_meta : dict
        A dictionary containing metadata for each segment.
        
    aggregates : List[str], optional
        A list of aggregation methods to apply to the quantification results.
        
    Returns
    -------
    dict
        A dictionary containing the aggregated quantification results for arm swing parameters.
    """
    arm_swing_parameters = [DataColumns.RANGE_OF_MOTION, DataColumns.PEAK_VELOCITY]

    uq_segment_cats = set([segment_meta[x][DataColumns.SEGMENT_CAT] for x in df_arm_swing_params[DataColumns.SEGMENT_NR].unique()])

    aggregated_results = {}
    for segment_cat in uq_segment_cats:
        cat_segments = [x for x in segment_meta.keys() if segment_meta[x][DataColumns.SEGMENT_CAT] == segment_cat]

        aggregated_results[segment_cat] = {
            'duration_s': sum([segment_meta[x]['duration_s'] for x in cat_segments])
        }

        df_arm_swing_params_cat = df_arm_swing_params[df_arm_swing_params[DataColumns.SEGMENT_NR].isin(cat_segments)]
        
        for arm_swing_parameter in arm_swing_parameters:
            for aggregate in aggregates:
                aggregated_results[segment_cat][f'{aggregate}_{arm_swing_parameter}'] = aggregate_parameter(df_arm_swing_params_cat[arm_swing_parameter], aggregate)

    aggregated_results['all_segment_categories'] = {
        'duration_s': sum([segment_meta[x]['duration_s'] for x in segment_meta.keys()])
    }

    for arm_swing_parameter in arm_swing_parameters:
        for aggregate in aggregates:
            aggregated_results['all_segment_categories'][f'{aggregate}_{arm_swing_parameter}'] = aggregate_parameter(df_arm_swing_params[arm_swing_parameter], aggregate)

    return aggregated_results


def extract_temporal_domain_features(
        config, 
        windowed_acc: np.ndarray, 
        windowed_grav: np.ndarray, 
        grav_stats: List[str] = ['mean']
        ) -> pd.DataFrame:
    """
    Compute temporal domain features for the accelerometer signal.

    This function calculates various statistical features for the gravity signal 
    and computes the standard deviation of the accelerometer's Euclidean norm.

    Parameters
    ----------
    config : object
        Configuration object containing the accelerometer and gravity column names.
    windowed_acc : numpy.ndarray
        A 2D numpy array of shape (N, M) where N is the number of windows and M is 
        the number of accelerometer values per window.
    windowed_grav : numpy.ndarray
        A 2D numpy array of shape (N, M) where N is the number of windows and M is 
        the number of gravity signal values per window.
    grav_stats : list of str, optional
        A list of statistics to compute for the gravity signal (default is ['mean']).

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the computed features, with each row corresponding 
        to a window and each column representing a specific feature.
    """
    # Compute gravity statistics (e.g., mean, std, etc.)
    feature_dict = {}
    for stat in grav_stats:
        stats_result = compute_statistics(data=windowed_grav, statistic=stat)
        for i, col in enumerate(config.gravity_cols):
            feature_dict[f'{col}_{stat}'] = stats_result[:, i]

    # Compute standard deviation of the Euclidean norm of the accelerometer signal
    feature_dict['accelerometer_std_norm'] = compute_std_euclidean_norm(data=windowed_acc)

    return pd.DataFrame(feature_dict)


def extract_spectral_domain_features(
        windowed_data: np.ndarray,
        config, 
        sensor: str, 
    ) -> pd.DataFrame:
    """
    Compute spectral domain features for a sensor's data.

    This function computes the periodogram, extracts power in specific frequency bands, 
    calculates the dominant frequency, and computes Mel-frequency cepstral coefficients (MFCCs) 
    for a given sensor's windowed data.

    Parameters
    ----------
    windowed_data : numpy.ndarray
        A 2D numpy array where each row corresponds to a window of sensor data.

    config : object
        Configuration object containing settings such as sampling frequency, window type, 
        frequency bands, and MFCC parameters.

    sensor : str
        The name of the sensor (e.g., 'accelerometer', 'gyroscope').
    
    Returns
    -------
    pd.DataFrame
        A DataFrame containing the computed spectral features, with each row corresponding 
        to a window and each column representing a specific feature.
    """
    # Initialize a dictionary to hold the results
    feature_dict = {}

    # Compute periodogram (power spectral density)
    freqs, psd = periodogram(
        x=windowed_data, 
        fs=config.sampling_frequency, 
        window=config.window_type, 
        axis=1
    )

    # Compute power in specified frequency bands
    for band_name, band_freqs in config.d_frequency_bandwidths.items():
        band_powers = compute_power_in_bandwidth(
            freqs=freqs,
            psd=psd, 
            fmin=band_freqs[0],
            fmax=band_freqs[1],
            include_max=False
        )
        for i, col in enumerate(config.axes):
            feature_dict[f'{sensor}_{col}_{band_name}'] = band_powers[:, i]

    # Compute dominant frequency for each axis
    dominant_frequencies = compute_dominant_frequency(
        freqs=freqs, 
        psd=psd, 
        fmin=config.spectrum_low_frequency, 
        fmax=config.spectrum_high_frequency
    )

    # Add dominant frequency features to the feature_dict
    for axis, freq in zip(config.axes, dominant_frequencies.T):
        feature_dict[f'{sensor}_{axis}_dominant_frequency'] = freq

    # Compute total power in the PSD
    total_power_psd = compute_total_power(psd)

    # Compute MFCCs
    mfccs = compute_mfccs(
        total_power_array=total_power_psd,
        config=config,
        multiplication_factor=4
    )

    # Combine the MFCCs into the features DataFrame
    mfcc_colnames = [f'{sensor}_mfcc_{x}' for x in range(1, config.mfcc_n_coefficients + 1)]
    for i, colname in enumerate(mfcc_colnames):
        feature_dict[colname] = mfccs[:, i]

    return pd.DataFrame(feature_dict)


import json
import numpy as np
import pandas as pd
import tsdf
from pathlib import Path
from scipy import signal
from scipy.interpolate import interp1d
from typing import List, Tuple, Union
from datetime import datetime

from paradigma.constants import TimeUnit, DataColumns
from paradigma.config import PPGConfig, IMUConfig
from paradigma.util import write_df_data, read_metadata, invert_watch_side


def resample_data(
    df: pd.DataFrame,
    time_column : str,
    values_column_names: List[str],
    sampling_frequency: int,
    resampling_frequency: int,
    tolerance: float | None = None
) -> pd.DataFrame:
    """
    Resamples sensor data to a specified frequency using cubic interpolation.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing the sensor data.
    time_column : str
        The name of the column containing the time data.
    values_column_names : List[str]
        A list of column names that should be resampled.
    sampling_frequency : int
        The original sampling frequency of the data (in Hz).
    resampling_frequency : int
        The frequency to which the data should be resampled (in Hz).
    tolerance : float, optional
        The tolerance added to the expected difference when checking 
        for contiguous timestamps. If not provided, it defaults to
        twice the expected interval.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the resampled data, where each column contains resampled values.
        The time column will reflect the new resampling frequency.

    Raises
    ------
    ValueError
        If the time array is not strictly increasing.

    Notes
    -----
    - Uses cubic interpolation for smooth resampling if there are enough points.
    - If only two timestamps are available, it falls back to linear interpolation.
    """
    # Set default tolerance if not provided to twice the expected interval
    if tolerance is None:
        tolerance = 2 * 1 / sampling_frequency

    # Extract time and values
    time_abs_array = np.array(df[time_column])
    values_array = np.array(df[values_column_names])

    # Ensure the time array is strictly increasing
    if not np.all(np.diff(time_abs_array) > 0):
        raise ValueError("Time array is not strictly increasing")
    
    # Ensure the time array is contiguous
    expected_interval = 1 / sampling_frequency
    timestamp_diffs = np.diff(time_abs_array)
    if np.any(np.abs(timestamp_diffs - expected_interval) > tolerance):
        raise ValueError("Time array is not contiguous")

    # Resample the time data using the specified frequency
    t_resampled = np.arange(time_abs_array[0], time_abs_array[-1], 1 / resampling_frequency)
    
    # Choose interpolation method
    interpolation_kind = "cubic" if len(time_abs_array) > 3 else "linear"
    interpolator = interp1d(time_abs_array, values_array, axis=0, kind=interpolation_kind, fill_value="extrapolate")
    
    # Interpolate
    resampled_values = interpolator(t_resampled)

    # Create a DataFrame with the resampled data
    df_resampled = pd.DataFrame(resampled_values, columns=values_column_names)
    df_resampled[time_column] = t_resampled

    # Return the DataFrame with columns in the correct order
    return df_resampled[[time_column] + values_column_names]


def butterworth_filter(
    data: np.ndarray,
    order: int,
    cutoff_frequency: Union[float, List[float]],
    passband: str,
    sampling_frequency: int,
):
    """
    Applies a Butterworth filter to 1D or 2D sensor data.

    This function applies a low-pass, high-pass, or band-pass Butterworth filter to the 
    input data. The filter is designed using the specified order, cutoff frequency, 
    and passband type. The function can handle both 1D and 2D data arrays.

    Parameters
    ----------
    data : np.ndarray
        The sensor data to be filtered. Can be 1D (e.g., a single signal) or 2D 
        (e.g., multi-axis sensor data).
    order : int
        The order of the Butterworth filter. Higher values result in a steeper roll-off.
    cutoff_frequency : float or List[float]
        The cutoff frequency (or frequencies) for the filter. For a low-pass or high-pass filter, 
        this is a single float. For a band-pass filter, this should be a list of two floats, 
        specifying the lower and upper cutoff frequencies.
    passband : str
        The type of passband to apply. Options are:
        - 'hp' : high-pass filter
        - 'lp' : low-pass filter
        - 'band' : band-pass filter
    sampling_frequency : int
        The sampling frequency of the data in Hz. This is used to normalize the cutoff frequency.

    Returns
    -------
    np.ndarray
        The filtered sensor data. The shape of the output is the same as the input data.

    Raises
    ------
    ValueError
        If the input data has more than two dimensions, or if an invalid passband is specified.

    Notes
    -----
    The function uses `scipy.signal.butter` to design the filter and `scipy.signal.sosfiltfilt`
    to apply it using second-order sections (SOS) to improve numerical stability.
    """
    # Design the filter using second-order sections (SOS)
    sos = signal.butter(
        N=order,
        Wn=cutoff_frequency,
        btype=passband,
        analog=False,
        fs=sampling_frequency,
        output="sos",
    )

    # Apply the filter to the data
    if data.ndim == 1:  # 1D data case
        return signal.sosfiltfilt(sos, data)
    elif data.ndim == 2:  # 2D data case
        return signal.sosfiltfilt(sos, data, axis=0)
    else:
        raise ValueError("Data must be either 1D or 2D.")

def preprocess_imu_data(df: pd.DataFrame, config: IMUConfig, sensor: str, watch_side: str) -> pd.DataFrame:
    """
    Preprocesses IMU data by resampling and applying filters.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing raw accelerometer and/or gyroscope data.
    config : IMUConfig
        Configuration object containing various settings, such as time column name, accelerometer and/or gyroscope columns,
        filter settings, and sampling frequency.
    sensor: str
        Name of the sensor data to be preprocessed. Must be one of:
        - "accelerometer": Preprocess accelerometer data only.
        - "gyroscope": Preprocess gyroscope data only.
        - "both": Preprocess both accelerometer and gyroscope data.
    watch_side: str
        The side of the watch where the data was collected. Must be one of:
        - "left": Data was collected from the left wrist.
        - "right": Data was collected from the right wrist.

    Returns
    -------
    pd.DataFrame
        The preprocessed accelerometer and or gyroscope data with the following transformations:
        - Resampled data at the specified frequency.
        - Filtered accelerometer data with high-pass and low-pass filtering applied.
    
    Notes
    -----
    - The function applies Butterworth filters to accelerometer data, both high-pass and low-pass.
    """

    # Extract sensor column
    if sensor == 'accelerometer':
        values_colnames = config.accelerometer_cols
    elif sensor == 'gyroscope':
        values_colnames = config.gyroscope_cols
    elif sensor == 'both':
        values_colnames = config.accelerometer_cols + config.gyroscope_cols
    else:
        raise('Sensor should be either accelerometer, gyroscope, or both')
        
    # Resample the data to the specified frequency
    df = resample_data(
        df=df,
        time_column=DataColumns.TIME,
        values_column_names=values_colnames,
        sampling_frequency=config.sampling_frequency,
        resampling_frequency=config.sampling_frequency
    )

    # Invert the IMU data if the watch was worn on the right wrist
    df = invert_watch_side(df, watch_side, sensor)
    
    if sensor in ['accelerometer', 'both']:
      
        # Extract accelerometer data for filtering
        accel_data = df[config.accelerometer_cols].values

        # Define filter configurations for high-pass and low-pass
        filter_renaming_configs = {
        "hp": {"result_columns": config.accelerometer_cols, "replace_original": True},
        "lp": {"result_columns": [f'{col}_grav' for col in config.accelerometer_cols], "replace_original": False},
        }

        # Apply filters in a loop
        for passband, filter_config in filter_renaming_configs.items():
            filtered_data = butterworth_filter(
            data=accel_data,
            order=config.filter_order,
            cutoff_frequency=config.lower_cutoff_frequency,
            passband=passband,
            sampling_frequency=config.sampling_frequency,
            )

            # Replace or add new columns based on configuration
            df[filter_config["result_columns"]] = filtered_data

        values_colnames += config.gravity_cols

    df = df[[DataColumns.TIME, *values_colnames]]

    return df


def preprocess_ppg_data(df_ppg: pd.DataFrame, df_acc: pd.DataFrame, ppg_config: PPGConfig, 
                        imu_config: IMUConfig, start_time_ppg: str, start_time_imu: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Preprocess PPG and IMU (accelerometer only) data by resampling, filtering, and aligning the data segments.

    Parameters
    ----------
    df_ppg : pd.DataFrame
        DataFrame containing PPG data.
    df_acc : pd.DataFrame
        DataFrame containing accelerometer from IMU data.
    ppg_config : PPGPreprocessingConfig
        Configuration object for PPG preprocessing.
    imu_config : IMUPreprocessingConfig
        Configuration object for IMU preprocessing.
    start_time_ppg : str
        iso8601 formatted start time of the PPG data.
    start_time_imu : str
        iso8601 formatted start time of the IMU data.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        Preprocessed PPG and IMU data as DataFrames.
    
    """

    # Extract overlapping segments
    df_ppg_overlapping, df_acc_overlapping = extract_overlapping_segments(df_ppg, df_acc, start_time_ppg, start_time_imu)
    
    # Resample accelerometer data
    df_acc_proc = resample_data(
        df=df_acc_overlapping,
        time_column=DataColumns.TIME,
        values_column_names = list(imu_config.d_channels_accelerometer.keys()),
        sampling_frequency=imu_config.sampling_frequency,
        resampling_frequency=imu_config.sampling_frequency
    )

    # Resample PPG data
    df_ppg_proc = resample_data(
        df=df_ppg_overlapping,
        time_column=DataColumns.TIME,
        values_column_names = list(ppg_config.d_channels_ppg.keys()),
        sampling_frequency=ppg_config.sampling_frequency,
        resampling_frequency=ppg_config.sampling_frequency
    )


    # Extract accelerometer data for filtering
    accel_data = df_acc_proc[imu_config.accelerometer_cols].values

    # Define filter configurations for high-pass and low-pass
    filter_renaming_configs = {
    "hp": {"result_columns": imu_config.accelerometer_cols, "replace_original": True}}

    # Apply filters in a loop
    for passband, filter_config in filter_renaming_configs.items():
        filtered_data = butterworth_filter(
        data=accel_data,
        order=imu_config.filter_order,
        cutoff_frequency=imu_config.lower_cutoff_frequency,
        passband=passband,
        sampling_frequency=imu_config.sampling_frequency,
        )

        # Replace or add new columns based on configuration
        df_acc_proc[filter_config["result_columns"]] = filtered_data
    
    # Extract accelerometer data for filtering
    ppg_data = df_ppg_proc[ppg_config.ppg_colname].values

    # Define filter configurations for high-pass and low-pass
    filter_renaming_configs = {
    "bandpass": {"result_columns": ppg_config.ppg_colname, "replace_original": True}}

    # Apply filters in a loop
    for passband, filter_config in filter_renaming_configs.items():
        filtered_data = butterworth_filter(
        data=ppg_data,
        order=ppg_config.filter_order,
        cutoff_frequency=[ppg_config.lower_cutoff_frequency, ppg_config.upper_cutoff_frequency],
        passband=passband,
        sampling_frequency=ppg_config.sampling_frequency,
        )

        # Replace or add new columns based on configuration
        df_ppg_proc[filter_config["result_columns"]] = filtered_data
    
    return df_ppg_proc, df_acc_proc




def extract_overlapping_segments(df_ppg: pd.DataFrame, df_acc: pd.DataFrame, start_time_ppg: str, start_time_acc: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Extract DataFrames with overlapping data segments between accelerometer (from the IMU) and PPG datasets based on their timestamps.

    Parameters
    ----------
    df_ppg : pd.DataFrame
        DataFrame containing PPG data.
    df_acc : pd.DataFrame
        DataFrame containing accelerometer data from the IMU.
    start_time_ppg : str
        iso8601 formatted start time of the PPG data.
    start_time_acc : str
        iso8601 formatted start time of the accelerometer data.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        DataFrames containing the overlapping segments (time and values) of PPG and accelerometer data.
    """
    # Convert start times to Unix timestamps
    datetime_ppg_start = datetime.fromisoformat(start_time_ppg.replace("Z", "+00:00"))
    start_unix_ppg = int(datetime_ppg_start.timestamp())
    datetime_acc_start = datetime.fromisoformat(start_time_acc.replace("Z", "+00:00"))
    start_acc_ppg = int(datetime_acc_start.timestamp())

    # Calculate the time in Unix timestamps for each dataset because the timestamps are relative to the start time
    ppg_time = df_ppg[DataColumns.TIME] + start_unix_ppg
    acc_time = df_acc[DataColumns.TIME] + start_acc_ppg

    # Determine the overlapping time interval
    start_time = max(ppg_time.iloc[0], acc_time.iloc[0])
    end_time = min(ppg_time.iloc[-1], acc_time.iloc[-1])

    # Extract indices for overlapping segments
    ppg_start_index = np.searchsorted(ppg_time, start_time, 'left')
    ppg_end_index = np.searchsorted(ppg_time, end_time, 'right') - 1
    acc_start_index = np.searchsorted(acc_time, start_time, 'left')
    acc_end_index = np.searchsorted(acc_time, end_time, 'right') - 1

    # Extract overlapping segments from DataFrames
    df_ppg_overlapping = df_ppg.iloc[ppg_start_index:ppg_end_index + 1]
    df_acc_overlapping = df_acc.iloc[acc_start_index:acc_end_index + 1]

    return df_ppg_overlapping, df_acc_overlapping


from dataclasses import dataclass


@dataclass(frozen=True)
class DataColumns():
    """
    Class containing the data channels in `tsdf`.
    """
    ACCELEROMETER_X : str = "accelerometer_x"
    ACCELEROMETER_Y : str = "accelerometer_y"
    ACCELEROMETER_Z : str = "accelerometer_z"
    GYROSCOPE_X : str = "gyroscope_x"
    GYROSCOPE_Y : str = "gyroscope_y"
    GYROSCOPE_Z : str = "gyroscope_z"
    PPG : str = "green"
    TIME : str = "time"
    SEGMENT_NR : str = "segment_nr"
    SEGMENT_CAT: str = "segment_category"

    # Gait 
    GRAV_ACCELEROMETER_X : str = "accelerometer_x_grav"
    GRAV_ACCELEROMETER_Y : str = "accelerometer_y_grav"
    GRAV_ACCELEROMETER_Z : str = "accelerometer_z_grav"
    PRED_GAIT_PROBA: str = "pred_gait_proba"
    PRED_GAIT : str = "pred_gait"
    PRED_NO_OTHER_ARM_ACTIVITY_PROBA: str = "pred_no_other_arm_activity_proba"
    PRED_NO_OTHER_ARM_ACTIVITY : str = "pred_no_other_arm_activity"
    ANGLE : str = "angle"
    VELOCITY : str = "velocity"
    DOMINANT_FREQUENCY: str = "dominant_frequency"
    RANGE_OF_MOTION: str = "range_of_motion"
    PEAK_VELOCITY: str = "peak_velocity"

    # The following are used in tremor analysis
    PRED_TREMOR_PROBA: str = "pred_tremor_proba"
    PRED_TREMOR_LOGREG : str = "pred_tremor_logreg"
    PRED_TREMOR_CHECKED : str = "pred_tremor_checked"
    PRED_ARM_AT_REST: str = "pred_arm_at_rest"

    # Constants for PPG features
    VARIANCE: str = "variance"
    MEAN: str = "mean"
    MEDIAN: str = "median"
    KURTOSIS: str = "kurtosis"
    SKEWNESS: str = "skewness"
    DOMINANT_FREQUENCY: str = "dominant_frequency"
    RELATIVE_POWER: str = "relative_power"
    SPECTRAL_ENTROPY: str = "spectral_entropy"
    SIGNAL_NOISE_RATIO: str = "signal_noise_ratio"
    SECOND_HIGHEST_PEAK: str = "second_highest_peak"
    POWER_RATIO: str = "power_ratio"

    # Constants for PPG SQA feature using accerometer data
    ACC_POWER_RATIO: str = "acc_power_ratio"

    # Constants for SQA classification
    PRED_SQA_PROBA: str = "pred_sqa_proba"
    PRED_SQA_ACC_LABEL: str = "pred_sqa_acc_label"
    PRED_SQA: str = "pred_sqa"

    # Constants for pulse rate
    PULSE_RATE: str = "pulse_rate"
    
@dataclass(frozen=True)
class DataUnits():
    """
    Class containing the data channel unit types in `tsdf`.
    """
    ACCELERATION: str = "m/s^2"
    """ The acceleration is in m/s^2. """
    
    ROTATION: str = "deg/s"
    """ The rotation is in degrees per second. """
    
    GRAVITY: str = "g"
    """ The acceleration due to gravity is in g. """
    
    POWER_SPECTRAL_DENSITY: str = "g^2/Hz"
    """ The power spectral density is in g^2/Hz. """
    
    FREQUENCY: str = "Hz"
    """ The frequency is in Hz. """

    NONE: str = "none"
    """ The data channel has no unit. """
    

@dataclass(frozen=True)
class TimeUnit():
    """
    Class containing the `time` channel unit types in `tsdf`.
    """
    RELATIVE_MS : str = "relative_ms"
    """ The time is relative to the start time in milliseconds. """
    RELATIVE_S : str = "relative_s"
    """ The time is relative to the start time in seconds. """
    ABSOLUTE_MS : str = "absolute_ms"
    """ The time is absolute in milliseconds. """
    ABSOLUTE_S : str = "absolute_s"
    """ The time is absolute in seconds. """
    DIFFERENCE_MS : str = "difference_ms"
    """ The time is the difference between consecutive samples in milliseconds. """
    DIFFERENCE_S : str = "difference_s"
    """ The time is the difference between consecutive samples in seconds. """

UNIX_TICKS_MS: int = 1000



import numpy as np
import pickle

from pathlib import Path
from sklearn.base import BaseEstimator
from typing import Any, Optional

class ClassifierPackage:
    def __init__(self, classifier: Optional[BaseEstimator] = None, 
                 threshold: Optional[float] = None, 
                 scaler: Optional[Any] = None):
        """
        Initialize the ClassifierPackage with a classifier, threshold, and scaler.

        Parameters
        ----------
        classifier
            The trained classifier.
        threshold : float
            The classification threshold.
        scaler
            The trained scaler (e.g., StandardScaler or MinMaxScaler).
        """
        self.classifier = classifier
        self.threshold = threshold
        self.scaler = scaler

    def transform_features(self, X) -> np.ndarray:
        """
        Transform the input features using the scaler.

        Parameters
        ----------
        X : np.ndarray
            The input features.

        Return
        ------
        np.ndarray
            The transformed features.
        """
        if not self.scaler:
            return X
        return self.scaler.transform(X)

    def predict_proba(self, X) -> float:
        """
        Make predictions using the classifier and apply the threshold.

        Parameters
        ----------
        X : np.ndarray
            The input features.

        Return
        ------
        float
            The predicted probability.

        """
        if not self.classifier:
            raise ValueError("Classifier is not loaded.")
        return self.classifier.predict_proba(X)[:, 1]
    
    def predict(self, X) -> int:
        """
        Make predictions using the classifier and apply the threshold.

        Parameters
        ----------
        X : np.ndarray
            The input features.

        Return
        ------
        int
            The predicted class.

        """
        if not self.classifier:
            raise ValueError("Classifier is not loaded.")
        return int(self.predict_proba(X) >= self.threshold)
    
    def save(self, filepath: str | Path) -> None:
        """
        Save the ClassifierPackage to a file.

        Parameters
        ----------
        filepath : str
            The path to the file.
        """
        with open(filepath, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, filepath: str | Path):
        """
        Load a ClassifierPackage from a file.

        Parameters
        ----------
        filepath : str
            The path to the file.

        Return
        ------
        ClassifierPackage
            The loaded classifier package.
        """
        try:
            with open(filepath, 'rb') as f:
                return pickle.load(f)
        except Exception as e:
            raise ValueError(f"Failed to load classifier package: {e}") from e